"""
Navigate through slices in 3D image

Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# Too many instance attributes
# pylint: disable=R0902

from PyQt5 import QtWidgets
from PyQt5 import QtCore

class SliceNav(QtWidgets.QWidget):
    """ Navigate through slices in 3D image """
    changed = QtCore.pyqtSignal()

    def __init__(self, min_slice=0, max_slice=100):
        super().__init__()
        self.min_slice_idx = min_slice
        self.max_slice_idx = max_slice
        self.slice_idx = self.min_slice_idx
        self.init_ui()

    def update_range(self, new_image, mode):
        """ update range of slices based on shape of input image
            and view mode """
        if mode == 'axial':
            slice_count = new_image.shape[0]
        elif mode == 'coronal':
            slice_count = new_image.shape[1]
        elif mode == 'sagittal':
            slice_count = new_image.shape[2]
        else: 
            raise Exception(f"Unhandled mode:{mode}")

        self.max_slice_idx = slice_count - 1
        self.slider.setMaximum(self.max_slice_idx)

        # default to central slice.
        self.slice_idx = self.max_slice_idx // 2

        if self.slice_idx > self.max_slice_idx:
            self.slice_idx = self.max_slice_idx
            self.slider.setValue(self.slice_idx)

        self.slider.setValue(self.slice_idx)
        self.update_text()

    def update_text(self):
        """ Update label text. Can be useful when slice changes """
        # self.value_label.setText(f"{self.slice_idx+1}/{self.max_slice_idx+1}")
        self.value_label.setText(f"{self.slice_idx+1}")

    def init_ui(self):
        """ Create UI elements """
        self.layout = QtWidgets.QVBoxLayout()
        # self.label = QtWidgets.QLabel("Axial Slice")
        # self.label.setAlignment(QtCore.Qt.AlignCenter)
        # self.layout.addWidget(self.label)

        self.debounce = QtCore.QTimer()
        self.debounce.setInterval(5)
        self.debounce.setSingleShot(True)
        self.debounce.timeout.connect(self.debounced)

        self.slider_container = QtWidgets.QWidget()
        self.slider_layout = QtWidgets.QVBoxLayout()
        # self.slider_layout.addWidget(QtWidgets.QLabel("Axial"))
        self.slider_container.setLayout(self.slider_layout)
        self.slider = QtWidgets.QSlider(QtCore.Qt.Vertical)
        self.slider.setMinimum(self.min_slice_idx)
        self.slider.setMaximum(self.max_slice_idx)
        self.slider.setValue(self.slice_idx)
        self.slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.slider.setTickInterval(1)
        #self.slider.setFixedHeight(600)
        self.slider.valueChanged.connect(self.value_changed)
        self.slider_layout.addWidget(self.slider)
        self.value_label = QtWidgets.QLabel()
        self.update_text()
        #self.value_label.setText(f"{self.slice_idx+1}/{self.max_slice_idx+1}")
        self.value_label.setText(f"{self.slice_idx+1}")

        self.slider_layout.addWidget(self.value_label)
        self.layout.addWidget(self.slider_container)
        self.setLayout(self.layout)
        self.setFixedWidth(65)
        self.setWindowTitle("Axial Position")

    def value_changed(self):
        """ when value changed update slice index"""
        self.slice_idx = self.slider.value()
        self.update_text()
        self.debounce.start()

    def up_slice(self):
        self.slice_idx = self.slice_idx + 1
        if self.slice_idx > self.max_slice_idx:
            self.slice_idx = self.max_slice_idx
        self.slider.setValue(self.slice_idx)
        self.update_text()
        self.debounce.start()

    def down_slice(self):
        self.slice_idx = self.slice_idx - 1
        if self.slice_idx < 1:
            self.slice_idx = 1
        self.slider.setValue(self.slice_idx)
        self.update_text()
        self.debounce.start()


    def debounced(self):
        """ trigger event only so often """
        self.changed.emit()



"""
Copyright (C) 2020 Abraham Goerge Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=I1101,C0111,W0201,R0903,E0611, R0902, R0914
# pylint: disable=W0703 # Too broad an exception

import os
from os.path import splitext
import filecmp
import numpy as np
from skimage.io import imread
import nibabel as nib
import im_utils
from pathlib import Path


def penultimate_fname_with_segmentation(fnames, seg_dir):
    """
    Go through fnames and return the last one with a segmentation
    If no segmentations are found return None.
    """

    # if the segmentation folder contains directories
    # then we assume that each directory is for a class
    
    pen_fname = None
    last_fname = None
    seg_dirs = [os.path.join(seg_dir, d) for d
                in os.listdir(seg_dir) if os.path.isdir(os.path.join(seg_dir, d))]
    if not len(seg_dirs):
        seg_dirs = [seg_dir]
    
    seg_fnames = []

    for seg_dir in seg_dirs:
        seg_fnames += os.listdir(seg_dir)

    for fname in fnames:
       #=  fname.replace('.nrrd', '.nii.gz')
        if fname.endswith(
            ('.dcm', '.dicom', '.sr', '.DCM', '.DICOM', '.SR')):
            base_fname = fname.replace('.dcm', '.nii.gz')
        else:
            base_fname = fname.replace('.nrrd', '.nii.gz')
        if base_fname in seg_fnames:
            if last_fname is not None:
                pen_fname = last_fname
            last_fname = fname
    return pen_fname

def get_recursive_files(input_path):
    filenames = []
    path = Path(input_path)
    for p in path.rglob("*"):
        if os.path.isdir(str(p)) is False:
            filenames.append(p._str)
    return filenames


def get_annot_path(fname, train_dir, val_dir):
    """
    return path to annot if it is found in
    train or val annot dirs.
    Otherwise return None
    """
    if fname.endswith(
        ('.dcm', '.dicom', '.sr', '.DCM', '.DICOM', '.SR')):
      #  fname = fname.replace('.dcm', '.nii.gz')
        fname = fname.replace('.dcm', '.nii.gz').replace('.DCM', '.nii.gz')

    else:
        fname = fname.replace('.nrrd', '.nii.gz')
    train_path = os.path.join(train_dir, fname)
    val_path = os.path.join(val_dir, fname)
    if os.path.isfile(train_path):
        return train_path
    if os.path.isfile(val_path):
        return val_path
    return None


def get_new_annot_target_dir(train_annot_dir, val_annot_dir):
    """ Should we add new annotations to train or validation data? """
    #train_annots = os.listdir(train_annot_dir)
    #val_annots = os.listdir(val_annot_dir)
    train_annots = get_recursive_files(train_annot_dir)
    val_annots = get_recursive_files(val_annot_dir)
    
    
    train_annots = [f for f in train_annots if (splitext(f)[1] in ['.png', '.npy', '.gz'])]
    val_annots = [f for f in val_annots if (splitext(f)[1] in ['.png', '.npy', '.gz'])]

    num_train_annots = len(train_annots)
    num_val_annots = len(val_annots)
    
    if num_train_annots == 0 and num_val_annots == 0:
        # save in train directory first 
        return train_annot_dir
    # otherwise aim to get at least one annotation in train and validation.
    if num_train_annots == 0 and num_val_annots > 0:
        return train_annot_dir
    if num_train_annots > 0 and num_val_annots == 0:
        return val_annot_dir
    # then only add files to validation if there is at least 5x in train
    if num_train_annots >= (num_val_annots * 5):
        return val_annot_dir
    return train_annot_dir


def maybe_save_annotation_3d(image_data_shape, annot_data, annot_path,
                             fname, train_annot_dir, val_annot_dir, log):
    annot_data = annot_data.astype(np.byte)
    # if there is an existing annotation.
    if annot_path:
        existing_annot = im_utils.load_annot(annot_path,
                                             image_data_shape).astype(np.byte)
        # and the annot we are saving is different.
        if not np.array_equal(annot_data, existing_annot):
            # Then we must over-write the previously saved annoation.
            # The user is performing an edit, possibly correcting an error.
            # First save to project folder as temp file.
            # if the annotation is empty then delete the current annotation
            if np.sum(annot_data) == 0:
                log(f'maybe_save_annot,{fname},remove existing annotation as new data is empty')
                os.remove(annot_path)
            else:
                # otherwise update the data.
                log(f'maybe_save_annot,{fname},updated existing data')
                img = nib.Nifti1Image(annot_data, np.eye(4))
                img.to_filename(annot_path)
    else:
        # if there is not an existing annotation
        # and the annotation has some content
        if np.sum(annot_data) > 0:
            log(f'maybe_save_annot,{fname},create new')
            # then find the best place to put it based on current counts.
            annot_dir = get_new_annot_target_dir(train_annot_dir, val_annot_dir)
            # files starting with . are not used in training.
            tmp_annot_path = os.path.join(annot_dir, '.tmp_' + fname)
            dir_to_create = os.path.dirname(tmp_annot_path)
            Path(dir_to_create).mkdir(parents=True, exist_ok=True)
            annot_path = os.path.join(annot_dir, fname)
            dir_to_create = os.path.dirname(annot_path)
            Path(dir_to_create).mkdir(parents=True, exist_ok=True)
            
            img = nib.Nifti1Image(annot_data, np.eye(4))
            img.to_filename(tmp_annot_path) 
            # rename after finished saving to avoid error with loading partially saved annotation.
            os.rename(tmp_annot_path, annot_path)
        else:
            # if the annotation did not have content.
            # and there was not an existing annotation
            # then don't save anything, this data is useless for
            # training.
            log(f'maybe_save_annot,{fname},not saving as annotation data empty')

    return annot_path



"""
Handle training and segmentation for a specific project

Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# pylint: disable=W0511, E1136, C0111, R0902, R0914, W0703, R0913, R0915
# W0511 is TODO
import os
import time
import warnings
import traceback
from pathlib import Path
import json
import sys
from datetime import datetime
import copy
import random
import multiprocessing


import numpy as np
import torch
from torch.utils.data import DataLoader
import torch.nn.functional as F
from loss import get_batch_loss
from instructions import fix_config_paths

from datasets import RPDataset
from metrics import get_metrics, get_metrics_str, get_metric_csv_row
from model_utils import ensemble_segment_3d
from model_utils import create_first_model_with_random_weights
from model_utils import random_model
import model_utils
from model_utils import save_if_better, save_model
from metrics import metrics_from_val_tile_refs
import data_utils
#import patch_seg

from im_utils import is_image, load_image, save_then_move, save
import im_utils
from file_utils import ls


class Trainer():

    def __init__(self, sync_dir, config, ip=None, port=None):
        self.sync_dir = sync_dir
        self.ip = ip
        self.port = port
        self.patch_update_enabled = ip and port

        self.instruction_dir = os.path.join(self.sync_dir, 'instructions')
        self.training = False
        self.running = False
        self.train_set = None
        # Can be set by instructions.
        self.train_config = None
        self.model = None
        self.first_loop = True
        self.batch_size = 4 
        self.config = config
        self.num_workers = config['max_workers']
        print(self.num_workers, 'workers will be assigned for data loaders')
        self.optimizer = None
        self.val_tile_refs = []
        # used to check for updates
        self.annot_mtimes = []
        self.msg_dir = None
        self.epochs_without_progress = 0
        self.training_restart = False
        self.restart_best_val_dice = 0
        
        self.in_w = None
        self.out_w = None
        self.small_unet = config['small_unet']
        # approx 30 minutes
        self.max_epochs_without_progress = 60
        # These can be trigged by data sent from client
        self.valid_instructions = [self.start_training,
                                   self.segment,
                                   self.stop_training,
                                   self.segment_patch]

    def main_loop(self, on_epoch_end=None):
        print('Started main loop. Checking for instructions in',
              self.instruction_dir)
        if self.patch_update_enabled:
            print('start patch seg server')
            patch_seg.start_server(self.sync_dir, self.ip, self.port) # direct socket connection
            print('after start server')

        self.running = True
        while self.running:
            self.check_for_instructions()
            if self.training:
                # can take a while so checks for
                # new instructions are also made inside
                self.val_tile_refs = self.get_new_val_tiles_refs()
                if self.val_tile_refs:
                    # selected as it leads to training taking around 5x validation time
                    train_epoch_length = max(64, 2 * len(self.val_tile_refs))
                else:
                    train_epoch_length = 128
                epoch_result = self.one_epoch(self.model, 'train', length=train_epoch_length)
                if epoch_result:
                    (tps, fps, tns, fns) = epoch_result
                    train_m = get_metrics(np.sum(tps), np.sum(fps), np.sum(tns), np.sum(fns))
                    self.log_metrics('train', train_m)
                    print(get_metrics_str(train_m, to_use=['dice', 'precision', 'recall']))
            if self.training:
                self.validation()
                if on_epoch_end:
                    on_epoch_end()
            else:
                self.first_loop = True
                time.sleep(1.0)

    def add_config_shape(self, config):
        new_config = copy.deepcopy(config)
        num_classes = len(config['classes'])
        if self.in_w is None:
            in_w, out_w = model_utils.get_in_w_out_w_for_memory(num_classes, self.small_unet)
            self.in_w = in_w
            self.out_w = out_w
            print('found input width of', in_w, 'and output width of', out_w)
        new_config['in_w'] = self.in_w
        new_config['out_w'] = self.out_w
        new_config['in_d'] = 52
        new_config['out_d'] = 18
        return new_config


    def fix_config_paths(self, old_config):
        """ get paths relative to local machine """
        new_config = {}
        for k, v in old_config.items():
            if k == 'file_names':
                # names dont need a path appending
                new_config[k] = v
            elif k == 'classes':
                # classes should not be altered
                new_config[k] = v
            elif isinstance(v, list):
                # if its a list fix each string in the list.
                new_list = []
                for e in v:
                    new_val = e.replace('\\', '/')
                    new_val = os.path.join(self.sync_dir,
                                           os.path.normpath(new_val))
                    new_list.append(new_val)
                new_config[k] = new_list
            elif isinstance(v, str):
                v = v.replace('\\', '/')
                new_config[k] = os.path.join(self.sync_dir,
                                             os.path.normpath(v))
            else:
                new_config[k] = v
        return new_config

    def check_for_instructions(self):
        for fname in ls(self.instruction_dir):
            print('found instruction', fname)
            if self.execute_instruction(fname):
                os.remove(os.path.join(self.instruction_dir, fname))

    def execute_instruction(self, fname):
        fpath = os.path.join(self.instruction_dir, fname)
        name = fname.rpartition('_')[0] # remove hash
        if name in [i.__name__ for i in self.valid_instructions]:
            print('execute_instruction', name)
            try:
                with open(fpath, 'r') as json_file:
                    contents = json_file.read()
                    config = fix_config_paths(self.sync_dir, json.loads(contents))
                    getattr(self, name)(config)
            except Exception as e:
                tb = traceback.format_exc()
                print('Exception parsing instruction', e, tb)
                return False
        else:
            #TODO put in a log and display error to the user.
            raise Exception(f"unhandled instruction {name})")
        return True

    def stop_training(self, _):
        if self.training:
            self.training = False
            self.epochs_without_progress = 0
            message = 'Training stopped'
            self.write_message(message)
            self.log(message)

    def start_training(self, config):
        if self.training_restart:
            self.training_restart = False
            self.training = False

        if not self.training:
            self.train_config = config
            self.train_config['dataset_dir'] = self.config['dataset_folder']
            
            classes = ['annotations']
            if 'classes' in self.train_config:
                classes = self.train_config['classes']
            else:
                self.train_config['classes'] = classes

            self.val_tile_refs = [] # dont want to cache these between projects
            self.epochs_without_progress = 0
            self.msg_dir = self.train_config['message_dir']
            model_dir = self.train_config['model_dir']
            classes = self.train_config['classes']
            self.train_config = self.add_config_shape(self.train_config)

            model_paths = model_utils.get_latest_model_paths(model_dir, 1)
            if model_paths:
                self.model = model_utils.load_model(model_paths[0], classes,
                                                    self.small_unet)
            else:
                self.model = create_first_model_with_random_weights(
                    model_dir, classes, self.small_unet)
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=0.01,
                                             momentum=0.99, nesterov=True)

            self.model.train()
            self.training = True


    def reset_progress_if_annots_changed(self):
        train_annot_dirs = self.train_config['train_annot_dirs']
        val_annot_dirs = self.train_config['val_annot_dirs']
        new_annot_mtimes = []
        for annot_dir in train_annot_dirs + val_annot_dirs:
            for fname in ls(annot_dir):
                fpath = os.path.join(annot_dir, fname)
                new_annot_mtimes.append(os.path.getmtime(fpath))
        new_annot_mtimes = sorted(new_annot_mtimes)
        if new_annot_mtimes != self.annot_mtimes:
            print('reset epochs without progress as annotations have changed')
            self.epochs_without_progress = 0
        self.annot_mtimes = new_annot_mtimes

    def write_message(self, message):
        """ write a message for the user (client) """
        Path(os.path.join(self.msg_dir, message)).touch()


    def train_annotation_exists(self):
        train_annot_dirs = self.train_config['train_annot_dirs']
        found_train_annot = False
        for d in train_annot_dirs:
            if [is_image(a) for a in ls(d)]:
                found_train_annot = True
        return found_train_annot 

    def one_epoch(self, model, mode='train', val_tile_refs=None, length=None):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        if not self.train_annotation_exists():
            # no training until data ready
            return False

        if self.first_loop:
            self.first_loop = False
            self.write_message('Training started')
            self.log('Starting Training')

        if mode == 'val':
            dataset = RPDataset(self.train_config['val_annot_dirs'],
                                None,
                                self.train_config['dataset_dir'],
                                self.train_config['in_w'],
                                self.train_config['out_w'],
                                self.train_config['in_d'],
                                self.train_config['out_d'],
                                'val',
                                val_tile_refs)
            torch.set_grad_enabled(False)
            loader = DataLoader(dataset, self.batch_size * 2, shuffle=True,
                                collate_fn=data_utils.collate_fn,
                                num_workers=self.num_workers,
                                drop_last=False, pin_memory=True)
        elif mode == 'train':
            dataset = RPDataset(self.train_config['train_annot_dirs'],
                                self.train_config['train_seg_dirs'],
                                self.train_config['dataset_dir'],
                                self.train_config['in_w'],
                                self.train_config['out_w'],
                                self.train_config['in_d'],
                                self.train_config['out_d'],
                                'train',
                                val_tile_refs,
                                length=length)
            torch.set_grad_enabled(True)
            loader = DataLoader(dataset, self.batch_size, shuffle=False,
                                collate_fn=data_utils.collate_fn,
                                num_workers=self.num_workers,
                                drop_last=False, pin_memory=True)
            model.train()
        else:
            raise Exception(f"Invalid mode: {mode}")

        epoch_start = time.time()

        tps = []    
        fps = []
        tns = []
        fns = []
        loss_sum = 0
        for step, (batch_im_tiles, batch_fg_tiles,
                   batch_bg_tiles, batch_seg_tiles, batch_classes) in enumerate(loader):

            self.check_for_instructions()
            batch_im_tiles = torch.from_numpy(np.array(batch_im_tiles)).to(device)
            self.optimizer.zero_grad()
        
            # padd channels to allow annotation input (or not)
            # l,r, l,r, but from end to start    w  w  h  h  d  d, c, c, b, b
            if self.patch_update_enabled:
                model_input = F.pad(batch_im_tiles, (0, 0, 0, 0, 0, 0, 0, 2), 'constant', 0)
            else:
                model_input = batch_im_tiles
    
            # model_input[:, 0] is the input image
            # model_input[:, 1] is fg
            # model_input[:, 2] is bg
            if self.patch_update_enabled and mode == 'train':
                for i, (fg_tiles, bg_tiles) in enumerate(zip(batch_fg_tiles, batch_bg_tiles)):
                    # if it's trianing then with 50% chance 
                    # add the annotations to the model input
                    # Validation should not have access to the annotations.
                    if random.random() > 0.5:
                        # go through fg tiles and bg_tiles for each batch item
                        # in this case we know there is always 1 bg and 1 fg tile.
                        # at random add the annotation slice
                        for slice_idx in range(fg_tiles[0].shape[0]):
                            if torch.any(fg_tiles[0][slice_idx]) or torch.any(bg_tiles[0][slice_idx]):
                                # each slice with annotation is included with 50 percent probability.
                                # This allows the network to learn how to use the annotation to improve predictions
                                if random.random() > 0.5: 
                                    model_input[i, 1, slice_idx] = fg_tiles[0][slice_idx]
                                    model_input[i, 2, slice_idx] = bg_tiles[0][slice_idx]

            outputs = model(model_input)

            (batch_loss, batch_tps, batch_tns,
             batch_fps, batch_fns) = get_batch_loss(
                 outputs, batch_fg_tiles, batch_bg_tiles, batch_seg_tiles,
                 #outputs, batch_fg_tiles, batch_bg_tiles,
                 batch_classes, self.train_config['classes'],
                 compute_loss=(mode=='train'))

            tps += batch_tps
            fps += batch_fps
            tns += batch_tns
            fns += batch_fns

            if mode == 'train':
                loss_sum += batch_loss.item() # float
                batch_loss.backward()
                self.optimizer.step()

            if mode == 'train':
                sys.stdout.write(f"{mode} {(step+1) * self.batch_size}/"
                                 f"{len(loader.dataset)} "
                                 f" loss={round(batch_loss.item(), 3)} \r")
                sys.stdout.flush()

            self.check_for_instructions() # could update training parameter
            if not self.training: # in this context we consider validation part of training.
                return

        duration = round(time.time() - epoch_start, 3)
        print(f'{mode} epoch duration', duration,
              'time per instance', round((time.time() - epoch_start) / len(tps), 3))

        return [tps, fps, tns, fns]

    def assign_metrics_to_refs(self, tps, fps, tns, fns):
        # now go through and assign the errors to the appropriate tile refs.
        for i, (tp, fp, tn, fn) in enumerate(zip(tps, fps, tns, fns)):
            # go through the val tile refs to find the equivalent tile ref
            self.val_tile_refs[i][3] = [tp, fp, tn, fn]

    def get_new_val_tiles_refs(self):
        return im_utils.get_val_tile_refs(self.train_config['val_annot_dirs'],
                                          copy.deepcopy(self.val_tile_refs),
                                          out_shape=(self.train_config['out_d'],
                                                     self.train_config['out_w'],
                                                     self.train_config['out_w']))

    def log_metrics(self, name, metrics):
        fname = datetime.today().strftime('%Y-%m-%d')
        fname += f'_{name}.csv'
        fpath = os.path.join(self.train_config['log_dir'], fname)
        if not os.path.isfile(fpath):
            # write headers if file didn't exist
            print('date_time,true_positives,false_positives,true_negatives,'
                  'false_negatives,precision,recall,dice',
                  file=open(fpath, 'w+'))
        with open(fpath, 'a+') as log_file:
            log_file.write(get_metric_csv_row(metrics))
            log_file.flush()

    def validation(self):
        """ Get validation set loss for current model and previous model.
            log those metrics and update the model if the
            current model is better than the previous model.
            Also stop training if the current model hasnt
            beat the previous model for {max_epochs}
        """
        model_dir = self.train_config['model_dir']
        prev_model, prev_path = model_utils.get_prev_model(model_dir,
                                                           self.train_config['classes'],
                                                           self.small_unet)
        self.val_tile_refs = self.get_new_val_tiles_refs()

        if not self.val_tile_refs:
            # if we don't yet have any validation data
            # but we are training then just save the model
            # Assuming we will get better than random weights
            # by the time the user gets to the second / third image
            save_model(model_dir, self.model, prev_path)
            was_saved = True
        else:
            # for current model get errors for all tiles in the validation set.
            epoch_result = self.one_epoch(copy.deepcopy(self.model), 'val', self.val_tile_refs)
            if not epoch_result:
                # if we didn't get anything back then it means the
                # dataset did not contain any annotations so no need
                # to proceed with validation.
                return 
            (tps, fps, tns, fns) = epoch_result
        
            cur_m = get_metrics(np.sum(tps), np.sum(fps), np.sum(tns), np.sum(fns))
            self.log_metrics('cur_val', cur_m)

            prev_m = self.get_prev_model_metrics(prev_model)
            self.log_metrics('prev_val', prev_m)
            was_saved = save_if_better(model_dir, self.model, prev_path,
                                       cur_m['dice'], prev_m['dice'])
            if was_saved:
                # update the cache to use metrics from current model
                self.assign_metrics_to_refs(tps, fps, tns, fns)

        if was_saved:
            self.epochs_without_progress = 0
        else:
            self.epochs_without_progress += 1

        # if we are doing a restart (from random weights) then lets consider the
        # performance improvements local to that restart, 
        # rather than the best model from all starts.
        if self.training_restart: 
            # we know that data doesn't change during a restart 
            # (because this would cause the restart to stop and typical training to resumse)
            # so we keep track of the best dice so far for this specific restart, and extend
            # training if we beat it, i.e set epochs_without_progress to 0
            if cur_m['dice'] > self.restart_best_val_dice:
                print('local restart dice improvement from',
                      round(self.restart_best_val_dice, 4), 'to', round(cur_m['dice'], 4))
                self.restart_best_val_dice = cur_m['dice']
                self.epochs_without_progress = 0

        self.reset_progress_if_annots_changed()

        message = (f'Training {self.epochs_without_progress}'
                   f' of max {self.max_epochs_without_progress}'
                   ' epochs without progress')
        self.write_message(message)
        if self.epochs_without_progress >= self.max_epochs_without_progress:
            message = (f'Training finished as {self.epochs_without_progress}'
                       ' epochs without progress')
            print(message)
            self.log(message)
            self.training = False
            self.write_message(message)
            self.restart_training()
    
    def restart_training(self):
        print('Restarting training from scratch')
        self.write_message('Restarting training from scratch')
        self.log('Restarting training from scratch')
        self.training_restart = True
        self.restart_best_val_dice = 0 # need to beat this or restart will stop after 60 epochs
        self.val_tile_refs = [] # dont want to cache these
        self.epochs_without_progress = 0
        self.model = random_model(self.train_config['classes'])
        self.optimizer = torch.optim.SGD(self.model.parameters(), lr=0.01,
                                         momentum=0.99, nesterov=True)
        self.model.train()
        self.training = True

    def log(self, message):
        with open(os.path.join(self.sync_dir, 'server_log.txt'), 'a+') as log_file:
            log_file.write(f"{datetime.now()}|{time.time()}|{message}\n")
            log_file.flush()

    def segment(self, segment_config):
        """
        Segment {file_names} from {dataset_dir} using {model_paths}
        and save to {seg_dir}.

        If model paths are not specified then use
        the latest model in {model_dir}.

        If no models are in {model_dir} then create a
        random weights model and use that.

        TODO: model saving is a counter-intuitve side effect,
        re-think project creation process to avoid this
        """
        in_dir = segment_config['dataset_dir']
        seg_dir = segment_config['seg_dir']

        segment_config = self.add_config_shape(segment_config)
        
        classes = segment_config['classes']
        if "file_names" in segment_config:
            fnames = segment_config['file_names']
            assert isinstance(fnames, list), type(fnames)
        else:
            # default to using all files in the directory if file_names is not specified.
            fnames = ls(in_dir)
            #ls_annotations(in_dir)
        # if model paths not specified use latest.
        if "model_paths" in segment_config:
            model_paths = segment_config['model_paths']
        else:
            model_dir = segment_config['model_dir']
            model_paths = model_utils.get_latest_model_paths(model_dir, 1)
            # if latest is not found then create a model with random weights
            # and use that.
            if not model_paths:
                create_first_model_with_random_weights(model_dir, classes, self.small_unet)
                model_paths = model_utils.get_latest_model_paths(model_dir, 1)
        if "overwrite" in segment_config:
            overwrite = segment_config['overwrite']
        else:
            overwrite = False

        start = time.time()

        for fname in fnames:
            self.segment_file(in_dir, seg_dir, fname,
                              model_paths,
                              in_w=segment_config['in_w'],
                              out_w=segment_config['out_w'],
                              in_d=segment_config['in_d'],
                              out_d=segment_config['out_d'],
                              classes=classes,
                              overwrite=overwrite)

        duration = time.time() - start
        print(f'Seconds to segment {len(fnames)} images: ', round(duration, 3))


    def get_prev_model_metrics(self, prev_model, use_cache=True):
        # for previous model get errors for all tiles which do not yet have metrics
        refs_to_compute = []

        if use_cache:
            # for each val tile
            for t in self.val_tile_refs:
                if t[3] is None:
                    refs_to_compute.append(t)
        else:
            refs_to_compute = self.val_tile_refs

        print('computing prev model metrics for ', len(refs_to_compute),
              'out of', len(self.val_tile_refs))
        # if it is missing metrics then add it to refs_to_compute
        # then compute the errors for these tile refs
        if refs_to_compute:
            
            (tps, fps, tns, fns) = self.one_epoch(prev_model, 'val', refs_to_compute)
            assert len(tps) == len(fps) == len(tns) == len(fns) == len(refs_to_compute)

            # now go through and assign the errors to the appropriate tile refs.
            for tp, fp, tn, fn, computed_ref in zip(tps, fps, tns, fns, refs_to_compute):
                # go through the val tile refs to find the equivalent tile ref
                for i, ref in enumerate(self.val_tile_refs):
                    if ref[0] == computed_ref[0] and ref[1] == computed_ref[1]:
                        if use_cache:
                            assert self.val_tile_refs[i][3] is None, self.val_tile_refs[i][3]
                            assert ref[3] is None
                        ref[3] = [tp, fp, tn, fn]
                        assert self.val_tile_refs[i][3] is not None

        prev_m = metrics_from_val_tile_refs(self.val_tile_refs)
        return prev_m

    def segment_patch(self, segment_config):
        raise Exception('feature disabled')
        patch_seg.segment_patch(segment_config)
        
    def get_in_w_and_out_w_for_image(self, im, in_w, out_w):
        """ the input image may be smaller than the default 
            patch size, so find a patch size where the output patch
            size will fit inside the image """

        im_depth, im_height, im_width = im.shape

        if out_w < im_width and out_w < im_height:
            return in_w, out_w
        
        for valid_in_w, valid_out_w in model_utils.get_in_w_out_w_pairs():
            if valid_out_w < im_width and valid_out_w < im_height and valid_out_w < out_w:
                return valid_in_w, valid_out_w

        raise Exception('cannot find patch size small enough for image with shape' + str(im.shape))

    def segment_file(self, in_dir, seg_dir, fname, model_paths,
                     in_w, out_w, in_d, out_d, classes, overwrite=False):

        # segmentations are always saved as .nii.gz
        out_paths = []
        if len(classes) > 1:
            for c in classes:
                out_fname = fname.replace('.nrrd', '.nii.gz') # output to nii.gz regardless of input format.
                out_paths.append(os.path.join(seg_dir, c, out_fname))
        else:
            # segment to nifty as they don't get loaded repeatedly in training.
            out_paths = [os.path.join(seg_dir, fname)]

        if not overwrite and all([os.path.isfile(out_path) for out_path in out_paths]):
            print(f'Skip because found existing segmentation files for {fname}')
            return

        fpath = os.path.join(in_dir, fname)

        if not os.path.isfile(fpath):
            raise Exception(f'Cannot segment as missing file {fpath}')
        try:
            im = load_image(fpath)
            # TODO: Consider removing thie soon
            #im = np.rot90(im, k=3)
            # im = np.moveaxis(im, -1, 0) # depth moved to beginning
            # reverse lr and ud
            im = im[::-1, :, ::-1]
        except Exception as e:
            # Could be temporary issues reading the image.
            # its ok just skip it.
            print('Exception loading', fpath, e)
            return
        seg_start = time.time()
        print('segment image, input shape = ', im.shape, datetime.now())

        seg_in_w, seg_out_w = self.get_in_w_and_out_w_for_image(im, in_w, out_w) 
        segmented = ensemble_segment_3d(model_paths, im, fname, self.batch_size,
                                        seg_in_w, seg_out_w, in_d, out_d, classes, self.small_unet)

        print(f'ensemble segment {fname}, dur', round(time.time() - seg_start, 2))
        
        for seg, outpath in zip(segmented, out_paths):
            # if the output folder doesn't exist then create it (class specific directory)
            out_dir = os.path.split(outpath)[0]
            if not os.path.exists(out_dir):
                print('making directory', out_dir)
                os.makedirs(out_dir)
            
            # catch warnings as low contrast is ok here.
            with warnings.catch_warnings():
                # create a version with alpha channel
                warnings.simplefilter("ignore")
                outpath = outpath.replace('.nrrd', '.nii.gz').replace('.DCM', '.nii.gz').replace('.dcm', '.nii.gz')
                save(outpath, seg)



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import time
from pathlib import Path


def get_recursive_files(input_path):
    filenames = []
    path = Path(input_path)
    for p in path.rglob("*"):
        if os.path.isdir(str(p)) is False:
            _, _, filename = p._str.partition(input_path + os.path.sep)
            filenames.append(filename)
    return filenames

def ls(dir_path):
    """
    list directory with
    retry as there may be temporary issues with a mounted network drive. 
    hidden files are not returned
    """
    retries = 100
    for _ in range(retries):
        try:
            fnames = os.listdir(dir_path)
            #fnames = get_recursive_files(dir_path)
            # Don't show hidden files
            # These can happen due to issues like file system 
            # synchonisation technology. RootPainter doesn't use them anywhere
            fnames = [f for f in fnames if f[0] != '.']
            return fnames
        except Exception as ex:
            print(f'exception listing file names from {dir_path}')
            print(ex)
            time.sleep(1)
    raise Exception(f'Cannot list file names from {dir_path} after {retries} retries')

def ls_annotations(dir_path):
    """
    list directory with
    retry as there may be temporary issues with a mounted network drive. 
    hidden files are not returned
    """
    retries = 100
    for _ in range(retries):
        try:
           # fnames = os.listdir(dir_path)
            fnames = get_recursive_files(dir_path)
            # Don't show hidden files
            # These can happen due to issues like file system 
            # synchonisation technology. RootPainter doesn't use them anywhere
            fnames = [f for f in fnames if f[0] != '.']
            return fnames
        except Exception as ex:
            print(f'exception listing file names from {dir_path}')
            print(ex)
            time.sleep(1)
    raise Exception(f'Cannot list file names from {dir_path} after {retries} retries')





"""
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from logging import root
import os
import time
import numpy as np
from functools import partial
from PyQt5 import QtCore
from enum import Enum


from tcp_utils import request_patch_seg, establish_connection
from instructions import fix_instruction_paths


class SegmentPatchThread(QtCore.QThread):

    complete = QtCore.pyqtSignal(np.ndarray)

    def __init__(self, annot_patch, content, server_ip, server_port):
         super().__init__()
         self.annot_patch = annot_patch
         self.content = content
         self.server_ip = server_ip
         self.server_port = server_port

    def run(self):
        # if running without data, just establish connection
        if self.annot_patch is None:
            establish_connection(self.server_ip, self.server_port)
        else:
            patch_seg = request_patch_seg(self.annot_patch,
                                          self.content)
            self.complete.emit(patch_seg)


class SegState(Enum):
    IDLE = 1
    SEGMENTING = 2 

class PatchSegmentor():
    """ 
    Holds the reference to root_painter
    and a reference to a QThread object that
    handles the server request to get the segmented patch.
    Then updates the root_painter segmentation data and 
    and triggers re-render.
    """

    def __init__(self, root_painter):
        self.root_painter = root_painter
        self.state = SegState.IDLE
        self.seg_patch_thread = SegmentPatchThread(None, None,
                                                   root_painter.server_ip,
                                                   root_painter.server_port)
        self.seg_patch_thread.start()


    def patch_received(self, patch_seg, fname):
        self.state = SegState.IDLE
        # only update if the image is still the same
        if self.root_painter.fname == fname:
            # TODO: I know 17 from memory. Compute '17' by 
            #       comparing input and output size
            seg_start_x = self.x_start + 17
            seg_start_y = self.y_start + 17
            seg_start_z = self.z_start + 17
            # first copy the seg data.
            new_seg_data = np.array(self.root_painter.seg_data)
            # update based on new patch data
            new_seg_data[
                seg_start_z:seg_start_z+patch_seg.shape[0],
                seg_start_y:seg_start_y+patch_seg.shape[1],
                seg_start_x:seg_start_x+patch_seg.shape[2]
            ] = patch_seg
            # now only update the displayed seg data based on the region the user intended to update
            self.root_painter.seg_data[
                self.z_valid_min:self.z_valid_max] = new_seg_data[self.z_valid_min:self.z_valid_max]
            for v in self.root_painter.viewers:
                if v.isVisible():
                    v.update_seg_slice()
                    v.update_outline()
            print('patch seg duration = ', time.time() - self.start_time)


    def segment_patch(self, centroid_x, centroid_y, z):
        """
        Prepare the data from the annotation and passes only the essentials.
        to a thread that uses a TCP SSL socket to send the patch details to the server
        and get back the patch segmentation.

        When the segmentation is received from the server, the patch_received event handler
        will be invoked.
        """
        if self.state == SegState.IDLE:
            self.state = SegState.SEGMENTING
            self.start_time = time.time()
            x = centroid_x
            y = centroid_y
            z = z
            root_painter = self.root_painter
            z = z + 1 # we dont segment the current slice. It gets confusing.
            # keep track of the region the user wanted to update. We dont allow changes outside of this region.
            self.z_valid_min = z
            self.z_valid_max = self.z_valid_min + root_painter.output_shape[0]

            # Move the centroid of the predicted region down by half the output depth
            # That way the whole output has a chance to be in the valid region.
            z = z + (root_painter.output_shape[0] // 2)

            # clip so it's at least output//2. This should allow the edge of the image to 
            # be resegmented.
            z = max((root_painter.output_shape[0] // 2), z)
            y = max((root_painter.output_shape[1] // 2), y)
            x = max((root_painter.output_shape[2] // 2), x)

            # image dimensions
            _, d, h, w = root_painter.annot_data.shape

            # also make sure we don't try to segment too far out on the other side
            z = min(d - (root_painter.output_shape[0] // 2), z)
            y = min(h - (root_painter.output_shape[1] // 2), y)
            x = min(w - (root_painter.output_shape[2] // 2), x)

            # and not more than the image shape
            # as only centroid is supplied, we must calculate left and right.
            self.z_start =  z - (root_painter.input_shape[0] // 2)
            z_end = self.z_start + root_painter.input_shape[0]
            self.y_start = y - (root_painter.input_shape[1] // 2)
            y_end = self.y_start + root_painter.input_shape[1]
            self.x_start = x - (root_painter.input_shape[2] // 2)
            x_end = self.x_start + root_painter.input_shape[2]

            content = {
                # required to know which image to load
                "file_name": root_painter.fname, 
                "dataset_dir": root_painter.dataset_dir,
                # required to know where to load the model from
                "model_dir": root_painter.model_dir,
                # which region of the image to segment
                "z_start": self.z_start, "z_end": z_end,
                "y_start": self.y_start, "y_end": y_end,
                "x_start": self.x_start, "x_end": x_end,
                # which classes to use with this model
                "classes": root_painter.classes
            }
            content = fix_instruction_paths(content, root_painter.sync_dir)
            # And then create an annotation for the patch input region
            # NOTE: annot_data shape is [2, d, h, w]
            #       First dimension is for bg (0) and fg (1)

            # if annotation goes outside the image, then we must pad it.
            pad_z_start = min(0, self.z_start) * -1
            pad_y_start = min(0, self.y_start) * -1
            pad_x_start = min(0, self.x_start) * -1 
            
            annot_patch = root_painter.annot_data[:,
                                                max(0, self.z_start):z_end,
                                                max(0, self.y_start):y_end,
                                                max(0, self.x_start):x_end]
            annot_patch = annot_patch.astype(bool)
            pad_z_end = root_painter.input_shape[0] - (annot_patch.shape[1] + pad_z_start)
            pad_y_end = root_painter.input_shape[1] - (pad_y_start + annot_patch.shape[2])
            pad_x_end = root_painter.input_shape[2] - (pad_x_start + annot_patch.shape[3])

            if sum([pad_z_start, pad_z_end, pad_y_start, pad_y_end, pad_x_start, pad_x_end]):
                annot_patch = np.pad(annot_patch, 
                                    [(0, 0), # dont add channels
                                    (pad_z_start, pad_z_end),
                                    (pad_y_start, pad_y_end),
                                    (pad_x_start, pad_x_end)],
                                    mode='constant')
            self.seg_patch_thread = SegmentPatchThread(annot_patch, content,
                                                    root_painter.server_ip,
                                                    root_painter.server_port)

            self.seg_patch_thread.complete.connect(partial(self.patch_received, fname=self.root_painter.fname))
            self.seg_patch_thread.start()


"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


Palette: Provides a way to add, edit and remove classes
"""

from PyQt5 import QtWidgets
from PyQt5 import QtCore

class BrushEditWidget(QtWidgets.QWidget):
    """
    Provide a way for a user to edit the name and colour of a brush
    """
    
    changed = QtCore.pyqtSignal()
    removed = QtCore.pyqtSignal()

    def __init__(self, name, show_remove):
        super().__init__()
        self.name = name
        self.initUI(show_remove)

    def initUI(self, show_remove):
        # Provide user with a way to edit the brush name
        self.layout = QtWidgets.QHBoxLayout()
        self.name_edit = QtWidgets.QLineEdit()
        self.name_edit.setText(self.name)
        self.name_edit.textChanged.connect(self.text_changed)
        self.layout.addWidget(self.name_edit)
    
        if show_remove:
            self.remove_btn = QtWidgets.QPushButton('Remove')
            self.remove_btn.clicked.connect(self.removed.emit)
            self.layout.addWidget(self.remove_btn)

        self.setLayout(self.layout)

    def text_changed(self):
        new_text = self.name_edit.text()
        self.name = new_text
        self.changed.emit()


class PaletteEditWidget(QtWidgets.QWidget):
    """ Add, edit and remove brushes """

    changed = QtCore.pyqtSignal()

    def __init__(self):
        super().__init__()
        self.brush_widgets = []
        self.initUI()


    def initUI(self):
        label = QtWidgets.QLabel()
        label.setText("Classes:")
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(label)

        # Use a container for the brushes so the add brush widget can go after.
        self.brushes_container = QtWidgets.QWidget()
        self.brushes_layout = QtWidgets.QVBoxLayout()
        self.brushes_container.setLayout(self.brushes_layout)
        self.layout.addWidget(self.brushes_container)

        # Default brush
        self.add_brush('foreground', show_remove=False)

        # btn to add brushes
        self.add_brush_btn = QtWidgets.QPushButton('Add class')
        self.add_brush_btn.clicked.connect(self.add_brush)
        self.layout.addWidget(self.add_brush_btn)

    def get_new_name(self):
        return f"class_{len(self.brush_widgets) + 1}"

    def add_brush(self, name=None, show_remove=True):
        if not name:
            name = self.get_new_name()

        brush = BrushEditWidget(name, show_remove)
        self.brush_widgets.append(brush)

        brush.removed.connect(self.remove_brush)
        self.brushes_layout.addWidget(brush)
        self.changed.emit()

    def remove_brush(self):
        brush = self.sender()
        self.brush_widgets.remove(brush)
        self.brushes_layout.removeWidget(brush)
        self.changed.emit()

    def get_brush_data(self):
        """ Used for saving the class names to JSON file """
        return [b.name for b in self.brush_widgets]



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# disable docstring warnings
#pylint: disable=C0111, C0111

# Module has no member
#pylint: disable=I1101

from PyQt5 import QtWidgets

class BaseProgressWidget(QtWidgets.QWidget):
    """
    Once a process starts this widget displays progress
    """
    def __init__(self, task):
        super().__init__()
        self.task = task
        self.initUI()

    def initUI(self):
        layout = QtWidgets.QVBoxLayout()
        self.layout = layout # to add progress bar later.
        self.setLayout(layout)
        # info label for user feedback
        info_label = QtWidgets.QLabel()
        info_label.setText("")
        layout.addWidget(info_label)
        self.info_label = info_label
        self.progress_bar = QtWidgets.QProgressBar(self)
        self.layout.addWidget(self.progress_bar)
        self.setWindowTitle(self.task)

    def onCountChanged(self, value, total):
        self.info_label.setText(f'{self.task} {value}/{total}')
        self.progress_bar.setValue(value)

    def done(self):
        QtWidgets.QMessageBox.about(self, self.task,
                                    f'{self.task} complete')
        self.close()



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=C0111, I1101, E0611
import os

from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5.QtCore import Qt
import json
from pathlib import Path, PurePosixPath
from im_utils import getDataFromDatabase

class NavWidget(QtWidgets.QWidget):
    """ Shows next and previous buttons as well as image position in folder.
    """
    file_change = QtCore.pyqtSignal(str)
    class_change = QtCore.pyqtSignal(str)


    def __init__(self, all_fnames, classes, before_change):
        super().__init__()
        self.image_path = None
        self.all_fnames = all_fnames
        # validation.
        self.before_change = before_change
        self.classes = classes
        self.settings = self.get_config()
        self.initUI()

    def get_config(self):
        settings_path = os.path.join(Path.home(), 'root_painter_settings.json')
        settings = json.load(open(settings_path, 'r'))
        return settings


    def initUI(self):
        # container goes full width to allow contents to be center aligned within it.
        nav = QtWidgets.QWidget()
        nav_layout = QtWidgets.QHBoxLayout()

        # && to escape it and show single &
        self.prev_image_button = QtWidgets.QPushButton('< Previous')
        self.prev_image_button.setFocusPolicy(Qt.NoFocus)
        self.prev_image_button.clicked.connect(self.show_prev_image)
        nav_layout.addWidget(self.prev_image_button)
        self.nav_label = QtWidgets.QLabel()
        nav_layout.addWidget(self.nav_label)

        # && to escape it and show single &
        self.next_image_button = QtWidgets.QPushButton('Save && Next >')
        self.next_image_button.setFocusPolicy(Qt.NoFocus)
        self.next_image_button.clicked.connect(self.show_next_image)
        nav_layout.addWidget(self.next_image_button)

        # left, top, right, bottom
        nav_layout.setContentsMargins(0, 0, 0, 5)
        nav.setLayout(nav_layout)
        nav.setMaximumWidth(600)

        # only need a way to change 
        if len(self.classes) > 1:
            self.cb = QtWidgets.QComboBox()
            self.cb.addItems(self.classes)
            self.cb.currentIndexChanged.connect(self.selection_change)
            nav_layout.addWidget(self.cb)

        container_layout = QtWidgets.QHBoxLayout()
        container_layout.setAlignment(Qt.AlignCenter)
        container_layout.addWidget(nav)
        self.setLayout(container_layout)
        container_layout.setContentsMargins(0, 0, 0, 0)
    
    def selection_change(self, _):
        self.class_change.emit(self.cb.currentText())

    def get_path_list(self, dir_path):
        all_files = self.all_fnames
        all_paths = [os.path.abspath(os.path.join(os.path.abspath(dir_path), a))
                     for a in all_files]
        return all_paths
    
    def get_paths(self):
        all_files = self.all_fnames
        all_paths = [
            os.path.join(self.settings['dataset_folder'], a) for a in all_files
            ]
        return all_paths

    def show_next_image(self):
        if self.before_change():
            self.update_to_next_image()

    def update_to_next_image(self):
        self.next_image_button.setEnabled(False)
        self.next_image_button.setText('Loading..')
        self.next_image_button.setEnabled(False)
        QtWidgets.QApplication.processEvents()
        dir_path, _ = os.path.split(self.image_path)
        #all_paths = self.get_path_list(dir_path)
        all_paths = self.get_paths()
        cur_idx = all_paths.index(self.image_path)
        next_idx = cur_idx + 1
        if next_idx >= len(all_paths):
            next_idx = 0
        self.image_path = all_paths[next_idx]
        self.file_change.emit(self.image_path)
        self.update_nav_label()

    def show_prev_image(self):
        dir_path, _ = os.path.split(self.image_path)
       # all_paths = self.get_path_list(dir_path)
        all_paths = self.get_paths()
        cur_idx = all_paths.index(self.image_path)
        next_idx = cur_idx - 1
        if next_idx <= 0:
            next_idx = 0
        self.image_path = all_paths[next_idx]
        self.file_change.emit(self.image_path)
        self.update_nav_label()

    def update_nav_label(self):
        #dir_path, _ = os.path.split(self.image_path)
        #dir_path = self.get_dir_path(self.image_path)
        #all_paths = self.get_path_list(dir_path)
        all_paths = self.get_paths()
        cur_idx = all_paths.index(self.image_path)
        self.nav_label.setText(f'{cur_idx + 1} / {len(all_paths)}')


    def get_dir_path(self, image_path):
        for fname in self.all_fnames:
            if fname in image_path:
                dir_path = image_path[0:-len(fname)]
                return dir_path
                


"""
Show image in either sagittal or axial view

Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Too many instance attributes
# pylint: disable=R0902
import sys

import qimage2ndarray
from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5.QtCore import Qt

from view_state import ViewState
from graphics_scene import GraphicsScene
from graphics_view import CustomGraphicsView
from slice_nav import SliceNav
from visibility_widget import VisibilityWidget

import im_utils
import menus


class ImViewer(QtWidgets.QWidget):
    """ Show image, annotation and segmentation data and allow interaction
        Contains graphics scene which is where the image will be drawn
    """

    def __init__(self, parent, view_mode):
        super().__init__()
        self.annot_visible = True
        self.seg_visible = False
        self.image_visible = True
        self.guide_image_visible = False
        self.outline_visible = False
        self.image_pixmap_holder = None
        self.guide_image_pixmap_holder = None
        self.seg_pixmap_holder = None
        self.blank_pixmap = None
        self.black_pixmap = None
        self.seg_pixmap = None
        self.annot_pixmap_holder = None
        self.cur_slice_idx = None
        self.im_width = None
        self.im_height = None
        self.brush_color = None
        self.mode = view_mode # axial or sagittal
        self.parent = parent # lazy event handling.
        self.setMouseTracking(True) # track mouse over to remove cursor from pixmap
        self.init_ui()

    def activate(self):
        self.activateWindow()
        # important so pan works on first click
        self.graphics_view.setFocus()

    def init_ui(self):
        """ Create and assign settings for graphical components """

        scene = GraphicsScene(self)
        self.scene = scene
        
        self.graphics_view = CustomGraphicsView()
        self.graphics_view.zoom_change.connect(self.update_cursor)
        self.graphics_view.leave.connect(self.scene.clear_cursor)
        self.graphics_view.enter.connect(self.activate)
        self.outer_layout = QtWidgets.QVBoxLayout()
        if self.mode == 'axial':
            self.setLayout(self.outer_layout)

        self.inner_container = QtWidgets.QWidget()
        self.inner_layout = QtWidgets.QHBoxLayout()
        self.inner_container.setLayout(self.inner_layout)
        self.outer_layout.addWidget(self.inner_container)

        self.graphics_view.setScene(scene)
        self.graphics_view.mouse_scroll_event.connect(self.mouse_scroll)

        # Required so graphics scene can track mouse up when mouse is not pressed
        self.graphics_view.setMouseTracking(True)

        # bottom bar
        bottom_bar = QtWidgets.QWidget()
        self.bottom_bar_layout = QtWidgets.QHBoxLayout()
        self.bottom_bar_layout.setSpacing(0)
        bottom_bar.setLayout(self.bottom_bar_layout)
        self.outer_layout.addWidget(bottom_bar)

        # slice nav
        self.slice_nav = SliceNav()
        self.inner_layout.addWidget(self.slice_nav)
        self.slice_nav.changed.connect(self.update_slice_index)
        self.inner_layout.addWidget(self.graphics_view)
        if self.mode == 'axial':
            self.vis_widget = VisibilityWidget(QtWidgets.QVBoxLayout, self,
                                               show_guide=hasattr(self.parent, 'guide_image_dir'))
            self.vis_widget.setMaximumWidth(200)
            # left, top, right, bottom
            self.bottom_bar_layout.setContentsMargins(20, 0, 20, 0)
        else:
            self.vis_widget = VisibilityWidget(QtWidgets.QHBoxLayout, self,
                                               show_guide=hasattr(self.parent, 'guide_image_dir'))

            self.vis_widget.setMaximumWidth(500)
            # left, top, right, bottom
            self.bottom_bar_layout.setContentsMargins(120, 0, 0, 10)
        self.vis_widget.setMinimumWidth(200)
      
        self.bottom_bar_layout.addWidget(self.vis_widget)

    def set_color(self, _event, color=None):
        """ change the brush to a new color """
        self.brush_color = color
        self.update_cursor()

    def seg_checkbox_change(self, state):
        """ set checkbox to specified state and update visibility if required """
        checked = (state == QtCore.Qt.Checked)
        if checked is not self.seg_visible:
            self.show_hide_seg()

    def annot_checkbox_change(self, state):
        """ set checkbox to specified state and update visibility if required """
        checked = (state == QtCore.Qt.Checked)
        if checked is not self.annot_visible:
            self.show_hide_annot()

    def im_checkbox_change(self, state):
        """ set checkbox to specified state and update visibility if required """
        checked = (state == QtCore.Qt.Checked)
        if checked is not self.image_visible:
            self.show_hide_image()

    def outline_checkbox_change(self, state):
        """ set checkbox to specified state and update visibility if required """
        checked = (state == QtCore.Qt.Checked)
        if checked is not self.outline_visible:
            self.show_hide_outline()

    def guide_checkbox_change(self, state):
        """ set checkbox to specified state and update visibility if required """
        checked = (state == QtCore.Qt.Checked)
        if checked is not self.guide_image_visible:
            self.show_hide_guide_image()

    def store_annot_slice(self):
        if self.scene.annot_pixmap:
            im_utils.store_annot_slice(self.scene.annot_pixmap,
                                       self.parent.annot_data,
                                       self.cur_slice_idx,
                                       self.mode)

    def update_slice_index(self):
        """ Render the new slice as the slice index may have changed """
        self.parent.log(f'update_slice,mode:{self.mode},'
                        f'view_state:{self.parent.view_state},'
                        f'new_slice:{self.slice_nav.slice_idx},'
                        f'zoom:{self.graphics_view.zoom}')

        # As we have likley moved to a new slice. The history needs erasing.
        # to avoid undo taking us back to prevous states of another slice.
        if len(self.scene.history) > 0:
            # before updating the slice idx, store current annotation information
            if hasattr(self.scene, 'annot_pixmap'):
                self.parent.annot_data = im_utils.store_annot_slice(self.scene.annot_pixmap,
                                                                    self.parent.annot_data,
                                                                    self.cur_slice_idx,
                                                                    self.mode)
            self.scene.history = []
            self.scene.redo_list = []

        self.cur_slice_idx = self.slice_nav.slice_idx
        
        # update image, seg and annot at current slice.
        self.update_image_slice()
        self.update_guide_image_slice()

        self.update_seg_slice()
        self.update_annot_slice()
        self.update_outline()

        self.scene.update_axial_slice_pos_indicator()

    def keyPressEvent(self, event):
        # if control key is pressed then hide the paint brush as the user wants to pan now.
        if event.key() == 16777249:
            self.scene.setFocus()
            self.graphics_view.setFocus()
            # need to manually set ctrl_press to true to tell the cursor how to look
            self.update_cursor(ctrl_press=True)

    def mouse_scroll(self, event):
        """ handle mouse scroll event. zoom, change slice or change brush size """
        scroll_up = event.angleDelta().y() > 0
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        alt_down = (modifiers & QtCore.Qt.AltModifier)
        shift_down = (modifiers & QtCore.Qt.ShiftModifier)
        ctrl_down = (modifiers & QtCore.Qt.ControlModifier)

        if alt_down or shift_down:
            # change by 10% (nearest int) or 1 (min)
            increment = max(1, int(round(self.scene.brush_size / 10)))
            if scroll_up:
                self.scene.brush_size += increment
            else:
                self.scene.brush_size -= increment
            self.scene.brush_size = max(1, self.scene.brush_size)
            self.parent.log(f'update_brush_size,brush_size:{self.scene.brush_size}')
            self.update_cursor()

        elif ctrl_down:
            if scroll_up:
                self.zoom_in()
            else:
                self.zoom_out()
        else:
            if scroll_up:
                self.slice_nav.slider.setValue(self.slice_nav.slice_idx + 1)
            else:
                self.slice_nav.slider.setValue(self.slice_nav.slice_idx - 1)

    def zoom_in(self):
        self.graphics_view.zoom *= 1.1
        self.parent.log(f'update_zoom_scroll,zoom:{self.graphics_view.zoom}')
        self.graphics_view.update_zoom()

    def zoom_out(self):
        self.graphics_view.zoom /= 1.1
        self.parent.log(f'update_zoom_scroll,zoom:{self.graphics_view.zoom}')
        self.graphics_view.update_zoom()

    def show_hide_image(self):
        """ show or hide the current image.
            Could be useful to help inspect the
            segmentation or annotation """
        if self.image_visible:
            self.image_pixmap_holder.setPixmap(self.black_pixmap)
            self.image_visible = False
        else:
            self.image_pixmap_holder.setPixmap(self.graphics_view.image)
            self.image_visible = True
        self.vis_widget.im_checkbox.setChecked(self.image_visible)

    def show_hide_guide_image(self):
        """ show or hide the guide image.
            guide image can be used to help the annotator by providing another modality to inspect etc """
        if hasattr(self.graphics_view, 'guide_image'):
            if self.guide_image_visible:
                self.guide_image_pixmap_holder.setPixmap(self.blank_pixmap)
                self.guide_image_visible = False
            else:
                self.guide_image_pixmap_holder.setPixmap(self.graphics_view.guide_image)
                self.guide_image_visible = True
            self.vis_widget.guide_image_checkbox.setChecked(self.guide_image_visible)

    def show_hide_annot(self):
        """ show or hide the current annotations.
            Could be useful to help inspect the background image
        """
        if self.annot_visible:
            self.annot_pixmap_holder.setPixmap(self.blank_pixmap)
            self.annot_visible = False
        else:
            self.scene.annot_pixmap_holder.setPixmap(self.scene.annot_pixmap)
            self.annot_visible = True
        self.vis_widget.annot_checkbox.setChecked(self.annot_visible)

    def show_hide_outline(self):
        """ show or hide the current outline.
            Could be useful to help inspect the
            segmentation or annotation """
        if self.outline_visible:
            self.scene.outline_pixmap_holder.setPixmap(self.blank_pixmap)
            self.outline_visible = False
        else:
            self.scene.outline_pixmap_holder.setPixmap(self.scene.outline_pixmap)
            self.outline_visible = True
        self.vis_widget.outline_checkbox.setChecked(self.outline_visible)

    def show_hide_seg(self):
        """ show or hide the current segmentation. """
        if self.seg_visible:
            self.seg_pixmap_holder.setPixmap(self.blank_pixmap)
            self.seg_visible = False
        else:
            self.seg_pixmap_holder.setPixmap(self.seg_pixmap)
            self.seg_visible = True
        self.vis_widget.seg_checkbox.setChecked(self.seg_visible)

    def update_image(self):
        """ show a new image, annotation and segmentation """
        self.slice_nav.update_range(self.parent.img_data, self.mode)
        # used for saving the edited annotation information
        # before changing slice
        self.cur_slice_idx = self.slice_nav.slice_idx

        # clear history when moving to a new image
        self.scene.history = []
        self.scene.redo_list = []

        # render image, seg and annot at current slice.
        self.update_image_slice()
        self.update_guide_image_slice()

        self.update_seg_slice()
        self.update_annot_slice()
        self.update_outline()

        # draw cursor showing what will be painted
        if self.scene.cursor_pixmap == None: 
            self.scene.cursor_pixmap = QtGui.QPixmap(self.im_width, self.im_height)
            self.scene.cursor_pixmap.fill(Qt.transparent)
            self.scene.cursor_pixmap_holder = self.scene.addPixmap(self.scene.cursor_pixmap)

    def fill_slice(self):
        self.scene.annot_pixmap = im_utils.fill_annot(self.scene.annot_pixmap)
        self.scene.annot_pixmap_holder.setPixmap(self.scene.annot_pixmap)
        # Store updated slice
        self.parent.annot_data = im_utils.store_annot_slice(self.scene.annot_pixmap,
                                                            self.parent.annot_data,
                                                            self.cur_slice_idx,
                                                            self.mode)
        self.scene.history.append(self.scene.annot_pixmap.copy())
        # update the other views
        self.parent.update_viewer_annot_slice()
        self.parent.update_viewer_outline()
    
    def update_image_slice(self):
        """ show the already loaded image we are looking at, at the specific
            slice the user has specified and using the specific
            user specified contrast settings.
        """
        img_slice = im_utils.get_slice(self.parent.img_data, self.slice_nav.slice_idx, self.mode)
        img_slice = im_utils.norm_slice(img_slice,
                                        self.parent.contrast_slider.min_value,
                                        self.parent.contrast_slider.max_value,
                                        self.parent.contrast_slider.brightness_value)
        q_image = qimage2ndarray.array2qimage(img_slice)
        image_pixmap = QtGui.QPixmap.fromImage(q_image)
        im_size = image_pixmap.size()
        im_width, im_height = im_size.width(), im_size.height()
        assert im_width > 0
        assert im_height > 0
        self.graphics_view.image = image_pixmap # for resize (and re-assignment after re-showing) later
        self.im_width = im_width
        self.im_height = im_height
        self.scene.setSceneRect(-15, -15, im_width+30, im_height+30)

        # Used to replace the segmentation or annotation when they are not visible.
        self.blank_pixmap = QtGui.QPixmap(self.im_width, self.im_height)
        self.blank_pixmap.fill(Qt.transparent)

        self.black_pixmap = QtGui.QPixmap(self.im_width, self.im_height)
        self.black_pixmap.fill(Qt.black)
        
        if self.image_pixmap_holder:
            self.image_pixmap_holder.setPixmap(image_pixmap)
        else:
            self.image_pixmap_holder = self.scene.addPixmap(image_pixmap)
            # interpolate the image pixels to make it look less pixelated
            self.image_pixmap_holder.setTransformationMode(Qt.SmoothTransformation)
        if not self.image_visible:
            self.image_pixmap_holder.setPixmap(self.black_pixmap)

    def mouseMoveEvent(self, event):
        if self.scene.cursor_shown:
            self.scene.cursor_pixmap.fill(Qt.transparent)
            self.scene.cursor_pixmap_holder.setPixmap(self.scene.cursor_pixmap)
            self.scene.cursor_shown = False

    def update_seg_slice(self):
        """ render the slice for the current index
            or show a loading message if it doesn't exist """
        if self.parent.seg_data is not None:
            seg_slice = im_utils.get_slice(self.parent.seg_data,
                                           self.slice_nav.slice_idx,
                                           self.mode)
            self.seg_pixmap = im_utils.seg_slice_to_pixmap(seg_slice)
            self.scene.seg_pixmap = self.seg_pixmap

        else:
            # otherwise use blank
            self.seg_pixmap = QtGui.QPixmap(self.im_width, self.im_height)
            self.seg_pixmap.fill(Qt.transparent)
            painter = QtGui.QPainter()
            painter.begin(self.seg_pixmap)
            font = QtGui.QFont()
            font.setPointSize(48)
            painter.setFont(font)
            painter.setPen(QtGui.QPen(QtGui.QColor(255, 255, 255)))
            painter.setBrush(QtGui.QBrush(QtGui.QColor(255, 255, 255), Qt.SolidPattern))
            if sys.platform == 'win32':
                # For some reason the text has a different size
                # and position on windows
                # so change the background rectangle also.
                painter.drawRect(0, 0, 657, 75)
            else:
                painter.drawRect(10, 10, 465, 55)
            painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 150)))
            painter.drawText(16, 51, 'Loading segmentation')
            painter.end()

        if self.seg_pixmap_holder:
            self.seg_pixmap_holder.setPixmap(self.seg_pixmap)
            self.scene.seg_pixmap = self.seg_pixmap
        else:
            self.seg_pixmap_holder = self.scene.addPixmap(self.seg_pixmap)
        if not self.seg_visible:
            self.seg_pixmap_holder.setPixmap(self.blank_pixmap)


    def update_cursor(self, ctrl_press=False):
        """ render cursor based on zoom level,
            brush color and brush size """
        self.scene.update_cursor(ctrl_press)

        if self.parent.view_state == ViewState.ANNOTATING:
            brush_w = self.scene.brush_size * self.graphics_view.zoom * 0.93
            brush_w = round(max(brush_w, 3))
            canvas_w = round(max(brush_w, 30))
            pixel_map = QtGui.QPixmap(canvas_w, canvas_w)
            pixel_map.fill(Qt.transparent)
            painter = QtGui.QPainter(pixel_map)
            painter.drawPixmap(canvas_w, canvas_w, pixel_map)
            brush_rgb = self.brush_color.toRgb()
            painter.setPen(QtGui.QPen(QtGui.QColor(0, 0, 0, 180), 2,
                                      Qt.SolidLine, Qt.FlatCap))
            # Draw black to show where cursor is even when brush is small
            painter.drawLine(0, round(canvas_w/2), canvas_w*2, round(canvas_w/2))
            painter.drawLine(round(canvas_w/2), 0, round(canvas_w/2), canvas_w*2)
            painter.end()
            cursor = QtGui.QCursor(pixel_map)
            QtWidgets.QApplication.restoreOverrideCursor()
            self.setCursor(cursor)

    def update_annot_slice(self):
        """ Update the annotation the user views """
        # if it's 3d then get the slice from the loaded numpy array
        annot_slice = im_utils.get_slice(self.parent.annot_data,
                                         self.slice_nav.slice_idx, self.mode)
        self.scene.annot_pixmap = im_utils.annot_slice_to_pixmap(annot_slice)


        # if annot_pixmap_holder is ready then setPixmap
        if self.annot_pixmap_holder:
            self.annot_pixmap_holder.setPixmap(self.scene.annot_pixmap)
        else:
            # or create the holder with the loaded pixmap
            self.annot_pixmap_holder = self.scene.addPixmap(self.scene.annot_pixmap)
        self.scene.annot_pixmap_holder = self.annot_pixmap_holder

        # so we can go back to the first stage of history
        if not self.scene.history:
            self.scene.history.append(self.scene.annot_pixmap.copy())
        if not self.annot_visible:
            self.annot_pixmap_holder.setPixmap(self.blank_pixmap)

    def update_outline(self):
        """ outline includes the segmentation corrected by the annotation
            and then just a line around the outside """
        if self.parent.seg_data is not None:
            seg_slice = im_utils.get_slice(self.parent.seg_data,
                                           self.slice_nav.slice_idx,
                                           self.mode)
            annot_slice = im_utils.get_slice(self.parent.annot_data,
                                             self.slice_nav.slice_idx, self.mode)
            self.scene.outline_pixmap = im_utils.get_outline_pixmap(seg_slice, annot_slice)
        else:
            self.scene.outline_pixmap = self.blank_pixmap
        

        # if outline_pixmap_holder is ready then setPixmap
        if self.scene.outline_pixmap_holder:
            self.scene.outline_pixmap_holder.setPixmap(self.scene.outline_pixmap)
        else:
            # or create the holder with the loaded pixmap
            self.scene.outline_pixmap_holder = self.scene.addPixmap(self.scene.outline_pixmap)
            self.scene.outline_pixmap_holder.setTransformationMode(Qt.SmoothTransformation)

        if not self.outline_visible:
            self.scene.outline_pixmap_holder.setPixmap(self.blank_pixmap)

    def update_guide_image_slice(self):
        """ guide shows another image on top of the main image """
        if hasattr(self.parent, 'guide_img_data'):
            img_slice = im_utils.get_slice(self.parent.guide_img_data, self.slice_nav.slice_idx, self.mode)
            img_slice = im_utils.norm_slice(img_slice,
                                            self.parent.contrast_slider.min_value,
                                            self.parent.contrast_slider.max_value,
                                            self.parent.contrast_slider.brightness_value)
            q_image = qimage2ndarray.array2qimage(img_slice)
            guide_image_pixmap = QtGui.QPixmap.fromImage(q_image)
            self.graphics_view.guide_image = guide_image_pixmap # for re-assignment after re-showing

            if self.guide_image_pixmap_holder:
                self.guide_image_pixmap_holder.setPixmap(guide_image_pixmap)
            else:
                self.guide_image_pixmap_holder = self.scene.addPixmap(guide_image_pixmap)
                # interpolate the image pixels to make it look less pixelated
                self.guide_image_pixmap_holder.setTransformationMode(Qt.SmoothTransformation)
            if not self.guide_image_visible:
                self.guide_image_pixmap_holder.setPixmap(self.blank_pixmap)


class ImViewerWindow(QtWidgets.QMainWindow, ImViewer):
    """ seperate viewer that has a menu in its own window """

    def init_ui(self):
        """ Create and assign settings for graphical components """
        super().init_ui()
        menus.add_view_menu(self, self, self.menuBar())
        menus.add_edit_menu(self, self, self.menuBar(), skip_fill=self.mode == 'sagittal')
        menus.add_brush_menu(self, self.menuBar())

        # This widget is designed to be a standalone window
        # so a central widget should be assigned.
        self.container = QtWidgets.QWidget()
        self.setCentralWidget(self.container)
        self.container.setLayout(self.outer_layout)
        self.setWindowTitle("Not Approved for clinical use. RootPainter3D - " + self.mode + 'view')

    def set_to_right_half_screen(self): 
        screenShape = QtWidgets.QDesktopWidget().screenGeometry()
        w = screenShape.width() // 2
        h = screenShape.height()
        x = screenShape.width() // 2
        y = 0
        self.setGeometry(x, y, w, h)



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# disable docstring warnings
#pylint: disable=C0111, C0111

# Module has no member
#pylint: disable=I1101

import re
from PyQt5 import QtWidgets
from PyQt5 import QtCore

class NameEditWidget(QtWidgets.QWidget):

    changed = QtCore.pyqtSignal()

    def __init__(self, entity):
        super().__init__()
        self.name = None
        self.initUI(entity)

    def initUI(self, entity):
        name_widget_layout = QtWidgets.QHBoxLayout()
        self.setLayout(name_widget_layout)
        edit_name_label = QtWidgets.QLabel()
        edit_name_label.setText(f"{entity} name:")
        name_widget_layout.addWidget(edit_name_label)
        self.name_edit = QtWidgets.QLineEdit()
        self.name_edit.textChanged.connect(self.text_changed)
        name_widget_layout.addWidget(self.name_edit)

    def text_changed(self):
        new_text = re.sub(r'\W+', '', self.name_edit.text())
        self.name_edit.setText(new_text)
        self.name = new_text
        self.changed.emit()



"""
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import socket
import time
import ssl
import rp_annot as rpa
import numpy as np
import json
import os
from pathlib import Path
import zlib

conn = None # global connection object to be re-used.

def establish_connection(ip, port):
    global conn
    if conn is None:
        print('Establish connection')
        conn = False # prevent running again while establishing
        server_cert = os.path.join(Path.home(), 'root_painter_server.public_key')
        client_cert = os.path.join(Path.home(), 'root_painter_client.public_key')
        client_key = os.path.join(Path.home(), 'root_painter_client.private_key')
        context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=server_cert)
        context.load_cert_chain(certfile=client_cert, keyfile=client_key)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        conn = context.wrap_socket(sock, server_side=False, server_hostname=ip)
        conn.connect((ip, port))
        log_cert = False
        if log_cert:
            print(f"SSL established. Peer cert: {conn.getpeercert()}'")

def request_patch_seg(annot_patch, segment_config):
    global conn
    annot_shape = np.array(annot_patch.shape)
    shape_bytes = annot_shape.tobytes()
    annot_patch_1d = annot_patch.reshape(-1)
    message = shape_bytes + rpa.compress(annot_patch_1d)
    conn.sendall(message)
    conn.sendall(b'end')
    message = json.dumps(segment_config).encode()
    conn.sendall(message)
    conn.sendall(b'cfg')
    seg_buffer = b'' # segmentation received from server.
    seg_shape = annot_shape[1:]
    # 17 is known for the current architecture. 
    # At some point in the future, we will probably want to accomodate 
    # different differences between input and output.
    seg_shape[0] = seg_shape[0] - (17*2)
    seg_shape[1] = seg_shape[1] - (17*2)
    seg_shape[2] = seg_shape[2] - (17*2)
    
    while True:
        data = conn.recv(16)
        # server sends 'end' when message over.
        if b'end' == data:
            seg_buffer = zlib.decompress(seg_buffer)
            seg_1d = rpa.decompress(seg_buffer, np.prod(seg_shape))
            seg_buffer = b'' # segmentation received from server.
            return seg_1d.reshape(seg_shape).astype(int)
        else:
            seg_buffer += data
    # we could call conn.close() to close the connection but instead we will leave it open.




"""
Form a batch

Copyright (C) 2021 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

def collate_fn(batch):
    num_items = len(batch)
    im_tiles = []
    batch_fgs = []
    batch_bgs = []
    batch_segs = []
    batch_classes = []

    for i in range(num_items):
        item = batch[i]
        class_data = {}
        im_tiles.append(item[0])
        batch_fgs.append(item[1])
        batch_bgs.append(item[2])
        batch_segs.append(item[3])
        batch_classes.append(item[4])

    im_tiles = np.array(im_tiles)
    return im_tiles, batch_fgs, batch_bgs, batch_segs, batch_classes




"""
Check for required folders on startup and create if required

Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import os
from pathlib import Path
import json
import copy

def startup_setup(arguments, sync_dir=None):
    """
    1. arguments contains dict with setting.
    2. If the sync dir doesn't exist then create it.
    3. If the required sync dir subfolders don't exist then create them.
    """

    # Get sync dir from settings file.
    if os.path.isfile(arguments['sync_dir']):
        sync_dir = Path(json.load(open(arguments['sync_dir'], 'r'))['sync_dir'])
        sync_dir_abs = os.path.abspath(sync_dir)
    else:
        # only create config file if user did not specify the sync dir on the command line
        if not sync_dir:
            # Or if the settings file doesn't exist get a sync_dir
            # from the user and save it to a settings file.
            sync_dir = input("Please specify RootPainter3D sync directory")
        sync_dir = os.path.expanduser(sync_dir)
        sync_dir_abs = os.path.abspath(sync_dir)
        with open(arguments['sync_dir'], 'w') as json_file:
            del arguments['sync_dir']
            arguments['contrast_presets'] = {'Mediastinal': [-125, 250, 100]}
            content = {**arguments, **{"sync_dir": sync_dir_abs}}
            print(f'Writing {sync_dir_abs} to {sync_dir}')
            json.dump(content, json_file, indent=4)

    # If sync_dir doesn't exist then create it.
    if not os.path.isdir(sync_dir_abs):
        print('Creating', sync_dir_abs)
        os.mkdir(sync_dir_abs)

    # RootPainter requires some folders to run. If they aren't already
    # in the sync_dir then create them
    required_subfolders = ['projects', 'datasets', 'instructions']
    for subfolder in required_subfolders:
        subfolder_path = os.path.join(sync_dir_abs, subfolder)
        if not os.path.isdir(subfolder_path):
            print('Creating', subfolder_path)
            os.mkdir(subfolder_path)



"""
Copyright (C) 2019, 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import random
import math
import os
from pathlib import Path

import torch
import numpy as np
from skimage import img_as_float32
from torch.utils.data import Dataset

from im_utils import load_train_image_and_annot
import im_utils
# from monai.transforms import (
#     AddChanneld,
#     Compose,
#     LoadImaged,
#     RandAffined,
#     RepeatChanneld,
#     MapTransform,
#     RandFlipd,
#     RandGaussianNoised,
#     RandZoomd,
#     NormalizeIntensityd,
#     RandGaussianSmoothd,
#     RandScaleIntensityd,
#     RandFlipd,
#     RandCropByPosNegLabeld,
#     SpatialPadd)

def rnd():
    """ Give higher than random chance to select the edges """
    return max(0, min(1, (1.2 * random.random()) - 0.1))


class RPDataset(Dataset):
    def __init__(self, annot_dirs, train_seg_dirs, dataset_dir, in_w, out_w,
                 in_d, out_d, mode, tile_refs=None, length=None):
        """
        in_w and out_w are the tile size in pixels

        target_classes is a list of the possible output classes
            the position in the list is the index (target) to be predicted by
            the network in the output.
            The value of the elmenent is the rgba (int, int, int) used to draw this
            class in the annotation.

            When the data is 3D the raw channels (for each class)
            are saved and the RGB values are not necessary.
        """
        self.mode = mode
        self.in_w = in_w
        self.out_w = out_w
        self.in_d = in_d
        self.out_d = out_d
        self.annot_dirs = annot_dirs
        self.train_seg_dirs = train_seg_dirs
        self.dataset_dir = dataset_dir
        assert (tile_refs is None) or (length is None) and (length or tile_refs)
        # if tile_refs are defined then these will be used.
        self.tile_refs = tile_refs
        # other wise length will return the number of items
        self.length = length

    def __len__(self):
        if self.mode == 'val':
            return len(self.tile_refs)
        if self.tile_refs is not None:
            return len(self.tile_refs)
        return self.length

    def __getitem__(self, i):
        if self.mode == 'val':
            return self.get_val_item(self.tile_refs[i])
        if self.tile_refs is not None:
            return self.get_train_item(self.tile_refs)
        return self.get_train_item()

    def get_train_item(self, tile_ref=None):
        return self.get_train_item_3d(tile_ref)

    def get_random_tile_3d(self, annots, segs, image, fname):
        # this will find something eventually as we know
        # all annotation contain labels somewhere

        # Limits for possible sampling locations from image (based on size of image)
        depth_lim = image.shape[0] - self.in_d
        bottom_lim = image.shape[1] - self.in_w
        right_lim = image.shape[2] - self.in_w

        attempts = 0 
        warn_after_attempts = 100
        while True:
            attempts += 1
            x_in = math.floor(rnd() * right_lim)
            y_in = math.floor(rnd() * bottom_lim)
            z_in = math.floor(rnd() * depth_lim)

            annot_tiles = []
            seg_tiles = []
            for seg, annot in zip(segs, annots):
                # Get the corresponding region of the annotation after network crop
                annot_tiles.append(annot[:,
                                         z_in:z_in+self.in_d,
                                         y_in:y_in+self.in_w,
                                         x_in:x_in+self.in_w])
                if seg is None:
                    seg_tiles.append(None)
                else:
                    seg_tiles.append(seg[z_in:z_in+self.in_d,
                                         y_in:y_in+self.in_w,
                                         x_in:x_in+self.in_w])



            # we only want annotations with defiend regions in the output area.
            # Otherwise we will have nothing to update the loss.
            if np.any([np.any(a) for a in annot_tiles]):
                # ok we have some annotation for this
                # part of the image so let's return the patch.
                im_tile = image[z_in:z_in+self.in_d,
                                y_in:y_in+self.in_w,
                                x_in:x_in+self.in_w]

                return annot_tiles, seg_tiles, im_tile
            if attempts > warn_after_attempts:
                print(f'Warning {attempts} attempts to get random patch from {fname}')
                warn_after_attempts *= 10

    def one_hot_decode(self, img):
        res = []
        for i in range(0, img.shape[0]):
            res.append(np.expand_dims(img[0]*(i+1), 0))

        return res

    def get_data_all_annots(self, k,  data_dict, annot):
        labels = annot.shape[0]
        remaning = list(range(0, labels))
        remaning.remove(k)

        data_dict['annots'] = np.expand_dims(annot[k], 0)
        for i in remaning:
            data_dict['annots_' + str(i)] = np.expand_dims(annot[i], 0)
        return data_dict, remaning
    
    def get_remaing_keys(self, data_dict):
        annot_names = []
        for k in data_dict:
            if k.startswith('annots_'):
                annot_names.append(k)
        return annot_names
    
    def combine_dicts(self, data_dict, k, remaining):
        combined = []
        total = len([k] + remaining)
        for i in range(0, total):
            if i == k:
                combined.append(data_dict['annots'])
            else:
                combined.append(data_dict['annots_' + str(i)])
        return np.concatenate(combined, axis=0)

    def get_random_tile_3d_byPosNeg(self, annots, segs, image, fname):
        attempts = 0 
        warn_after_attempts = 100
        #image = np.concatenate((np.expand_dims(image, 0), np.expand_dims(image, 0)), axis=0)
        data_dict = dict()
        data_dict['image'] = np.expand_dims(image, 0)

        while True:
            attempts += 1
            annot_tiles = []
            seg_tiles = []
            for seg, annot in zip(segs, annots):
                # Get the corresponding region of the annotation after network crop
                data_dict['segs'] = np.expand_dims(seg, 0)
                k = random.randint(0, annot.shape[0]-1)
                data_dict, remaining = self.get_data_all_annots(k, data_dict, annot)
                annot_names = self.get_remaing_keys(data_dict)
                spatialpad = SpatialPadd(
                    keys=['image', 'annots', 'segs'] + annot_names,
                    spatial_size=[self.in_d,
                                  self.in_w,
                                  self.in_w])
                randCrop = RandCropByPosNegLabeld(
                    keys=['image', 'annots', 'segs'] + annot_names,
                    label_key='annots',
                    spatial_size=[self.in_d,
                                  self.in_w,
                                  self.in_w],
                    pos=1, neg=1, num_samples=1)

                transform = Compose([spatialpad, randCrop])
                data_dict_return = transform(data_dict)
                seg_tiles.append(data_dict_return[0]['segs'][0])
                annot_tiles.append(self.combine_dicts(
                    data_dict_return[0],
                    k,
                    remaining))
            if np.any([np.any(a) for a in annot_tiles]):
                im_tile = data_dict_return[0]['image'][0]
                return annot_tiles, seg_tiles, im_tile
            if attempts > warn_after_attempts:
                print(f'Warning {attempts} attempts to get random patch from {fname}')
                warn_after_attempts *= 10

    def apply_data_augmentation(self, im_tile, annot_tiles, seg_tiles):
        # wrap in dict for transforms
        data_dict = dict()
        data_dict['image'] = np.expand_dims(im_tile, 0)
        if seg_tiles[0] is not None:
            data_dict['segs'] = np.expand_dims(seg_tiles[0], 0)
        # unroll the list of annots into a dict
        annots_dict = self.unroll_annots_to_dict(annot_tiles)
        # get names for each annot
        # annot_names = annots_dict.keys()
        data_dict.update(annots_dict)
        self.feature_names = data_dict.keys()
        transforms_to_apply = self.get_train_transforms()
        data_dict_return = transforms_to_apply(data_dict)
        im_tile = data_dict_return['image'][0, :, :, :]
        if seg_tiles[0] is not None:
            seg_tiles = [data_dict_return['segs'][0, :, :, :]]
        annot_tiles = [self.transform_dict_of_annots_to_numpy(annots_dict)]
        return im_tile, annot_tiles, seg_tiles


    def get_train_item_3d(self, tile_ref):
        # When tile_ref is specified we use these coordinates to get
        # the input tile. Otherwise we will sample randomly
        if tile_ref:
            raise Exception('not using these')
            im_tile, foregrounds, backgrounds, classes = self.get_tile_from_ref_3d(tile_ref)
            # For now just return the tile. We plan to add augmentation here.
            return im_tile, foregrounds, backgrounds, classes

        (image, annots, segs, classes, fname) = load_train_image_and_annot(self.dataset_dir,
                                                                           self.train_seg_dirs,
                                                                           self.annot_dirs)
        
       # annot_tiles, seg_tiles, im_tile = self.get_random_tile_3d_byPosNeg(annots, segs, image, fname)
        annot_tiles, seg_tiles, im_tile = self.get_random_tile_3d(annots, segs, image, fname)
        im_tile, annot_tiles, seg_tiles = self.apply_data_augmentation(
            im_tile, annot_tiles, seg_tiles)
        
        

        im_tile = img_as_float32(im_tile)
        im_tile = im_utils.normalize_tile(im_tile)
        # ensure image is still 32 bit after normalisation.
        im_tile = im_tile.astype(np.float32)
        # need list of foregrounds and masks for all tiles.
        foregrounds = []
        backgrounds = []
        for annot_tile in annot_tiles:
            #annot tile shape is  (2, 18, 194, 194)
            foreground = np.array(annot_tile)[1]
            background = np.array(annot_tile)[0]
            foreground = foreground.astype(np.int64)
            foreground = torch.from_numpy(foreground)
            foregrounds.append(foreground)
            background = background.astype(np.int64)
            background = torch.from_numpy(background)
            backgrounds.append(background)

        im_tile = im_tile.astype(np.float32)
        
        # add dimension for input channel
        im_tile = np.expand_dims(im_tile, axis=0)
        assert len(backgrounds) == len(seg_tiles)
        return im_tile, foregrounds, backgrounds, seg_tiles, classes

    def unroll_annots_to_dict(self, annots):
        annot_dict = dict()
        for i in range(0, annots[0].shape[0]):
            annot_dict['annots_' + str(i)] = np.expand_dims(annots[0][i], 0)
        return annot_dict

    def transform_dict_of_annots_to_numpy(self, annot_dict):
        # loop through the dict annd stack the numpy arrays
        for idx, annot in enumerate(list(annot_dict.keys())):
            if idx == 0:
                annots = annot_dict[annot]
            else:
                annots = np.vstack((annots, annot_dict[annot]))
        return annots

    def get_modes_sampling(self):
        modes = []
        for i in self.feature_names:
            if i == 'image':
                modes.append('bilinear')
            else:
                modes.append('nearest')
        return modes

    def import_transforms(self):
        from monai.transforms import (
            AddChanneld,
            Compose,
            LoadImaged,
            RandAffined,
            RepeatChanneld,
            MapTransform,
            RandFlipd,
            RandGaussianNoised,
            RandZoomd,
            NormalizeIntensityd,
            RandGaussianSmoothd,
            RandScaleIntensityd,
            RandFlipd,
            RandCropByPosNegLabeld,
            SpatialPadd)
        return Compose, LoadImaged, RandAffined, RepeatChanneld, MapTransform, RandFlipd, RandGaussianNoised, RandZoomd, NormalizeIntensityd, RandGaussianSmoothd, RandScaleIntensityd, RandFlipd, RandCropByPosNegLabeld, SpatialPadd
    def get_train_transforms(self):
        modes = self.get_modes_sampling()
        Compose, LoadImaged, RandAffined, RepeatChanneld, MapTransform, RandFlipd, RandGaussianNoised, RandZoomd, NormalizeIntensityd, RandGaussianSmoothd, RandScaleIntensityd, RandFlipd, RandCropByPosNegLabeld, SpatialPadd = self.import_transforms()
        transforms = Compose([
            # translation
            RandAffined(
                    keys=self.feature_names,
                    mode=modes,
                    prob=0.2,
                    spatial_size=(self.in_d,
                                  self.in_w,
                                  self.in_w),
                    translate_range=(
                            int(0.5*self.in_d),
                            int(0.22*self.in_w),
                            int(0.22*self.in_w)),
                    padding_mode="zeros"),
            # scaling spatial
            RandAffined(
                    keys=self.feature_names,
                    mode=modes,
                    prob=0.2,
                    spatial_size=(self.in_d,
                                  self.in_w,
                                  self.in_w),
                    scale_range=(0, 0.15, 0.15),
                    padding_mode="zeros"),
            # scaling temporal
            RandZoomd(
                keys=self.feature_names,
                prob=0.2,
                min_zoom=(0.5, 1, 1),
                max_zoom=(1.5, 1, 1),
                mode=['nearest' for i in range(0, len(self.feature_names))]),
            # rotation
            RandAffined(
                keys=self.feature_names,
                mode=modes,
                prob=0.2,
                rotate_range=(0.17, 0, 0),
                spatial_size=(self.in_d,
                              self.in_w,
                              self.in_w),
                padding_mode="zeros"),
            # flipping
            RandFlipd(keys=self.feature_names, spatial_axis=[0], prob=0.2),
            RandFlipd(keys=self.feature_names, spatial_axis=[1], prob=0.2),
            RandFlipd(keys=self.feature_names, spatial_axis=[2], prob=0.2),
            # Sclae intensity
            RandScaleIntensityd(keys=["image"], factors=0.3, prob=0.15),
            # Gaussian noise
            RandGaussianNoised(keys=["image"], std=0.01, prob=0.15),
            # Gaussian smoothing
            RandGaussianSmoothd(
                keys=["image"],
                sigma_x=(0.0, 0.0),
                sigma_y=(0.5, 1.15),
                sigma_z=(0.5, 1.15),
                prob=0.15,
            ),
        ])

        return transforms
    def get_val_item(self, tile_ref):
        return self.get_tile_from_ref_3d(tile_ref)

    def get_tile_from_ref_3d(self, tile_ref):
        """ return image tile, annotation tile and mask
            for a given file name ans location specified
            in x,y,z relative to the annotation """

        fname, (tile_x, tile_y, tile_z), _, _ = tile_ref
        image_path = os.path.join(self.dataset_dir, fname)
        # image could have nrrd extension
        if not os.path.isfile(image_path):
            image_path = image_path.replace('.nii.gz', '.nrrd')
        image = im_utils.load_with_retry(im_utils.load_image, image_path)
        #  needs to be swapped to channels first and rotated etc
        # to be consistent with everything else.
        # todo: consider removing this soon.
        #image = np.rot90(image, k=3)
       # image = np.moveaxis(image, -1, 0) # depth moved to beginning
        # reverse lr and ud
        #image = image[::-1, :, ::-1]

        # pad so seg will be size of input image
        image = np.pad(image, ((17, 17), (17, 17), (17, 17)), mode='constant')

        classes = []
        foregrounds = []
        backgrounds = []
        annot_tiles = []

        for annot_dir in self.annot_dirs:
            annot_path = os.path.join(annot_dir, fname)

            annot = im_utils.load_with_retry(im_utils.load_image, annot_path)
            classes.append(Path(annot_dir).parts[-2])
            
            # pad to provide annotation at same size as input image.
            annot = np.pad(annot, ((0, 0), (17, 17), (17, 17), (17, 17)), mode='constant')
            # The x, y and z are in reference to the annotation tile before padding.
            annot_tile = annot[:,
                               tile_z:tile_z+self.in_d,
                               tile_y:tile_y+self.in_w,
                               tile_x:tile_x+self.in_w]

            assert annot_tile.shape[1:] == (self.in_d, self.in_w, self.in_w), (
                f" annot is {annot_tile.shape}, and "
                f"should be ({self.in_d},{self.in_w},{self.in_w})")

            #else: # not auto-complete
            #    if os.path.isfile(annot_path):
            #        # The x, y and z are in reference to the annotation tile before padding.
            #        annot_tile = annot[:,
            #                           tile_z:tile_z+self.out_d,
            #                           tile_y:tile_y+self.out_w,
            #                           tile_x:tile_x+self.out_w]

            annot_tiles.append(annot_tile)

        im_tile = image[tile_z:tile_z + self.in_d,
                        tile_y:tile_y + self.in_w,
                        tile_x:tile_x + self.in_w]
 
        assert im_tile.shape == (self.in_d, self.in_w, self.in_w), (
            f" shape is {im_tile.shape}")
        
        for annot_tile in annot_tiles:
            foreground = np.array(annot_tile)[1]
            background = np.array(annot_tile)[0]
            foreground = foreground.astype(np.int64)
            foreground = torch.from_numpy(foreground)
            foregrounds.append(foreground)
            background = background.astype(np.int64)
            background = torch.from_numpy(background)
            backgrounds.append(background)

        im_tile = img_as_float32(im_tile)
        im_tile = im_utils.normalize_tile(im_tile)
        im_tile = im_tile.astype(np.float32)
        im_tile = np.expand_dims(im_tile, axis=0)
        segs = [None] * len(backgrounds)
        return im_tile, foregrounds, backgrounds, segs, classes



"""
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

def segment_full_image(root_painter, fname=None, overwrite=False):
    # if file name not specified then segment the current image.
    if fname is None:
        fname = root_painter.fname
    # send instruction to segment the image.
    root_painter.send_instruction('segment', {
        "dataset_dir": str(root_painter.dataset_dir),
        "seg_dir": root_painter.seg_dir,
        "file_names": [fname],
        "message_dir": root_painter.message_dir,
        "model_dir": root_painter.model_dir,
        "overwrite": overwrite,
        "classes": root_painter.classes # used for saving segmentation output to correct directories
    })




from setuptools import setup, find_packages

setup(
    name="cagpaint",
    version="0.1.7",
    packages=find_packages(),
    install_requires=[
        # Add your package dependencies here, e.g.,
        # 'numpy>=1.20.0',
    ],
    author="ChristianEschen",
    author_email="christian_eschen@hotmail.com",
    description="cagpaint",
    url="https://github.com/yourusername/your-package"
)



# pylint: disable=C0111, W0221, R0902
"""
U-Net architecture based on:
https://arxiv.org/abs/1505.04597
And modified to use Group Normalization
https://arxiv.org/abs/1803.08494
And then modified to use residual style connections.
With other alterations.

Copyright (C) 2019, 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import torch.nn as nn


class DownBlock(nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.pool = nn.MaxPool3d(2)
        self.conv1 = nn.Sequential(
            nn.Conv3d(in_channels, in_channels*2,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(32, in_channels*2)
        )
        self.conv2 = nn.Sequential(
            nn.Conv3d(in_channels*2, in_channels*2,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(32, in_channels*2)
        )
        self.conv1x1 = nn.Sequential(
            # down sample channels again.
            nn.Conv3d(in_channels*2, in_channels,
                      kernel_size=1, stride=1, bias=False)
        )

    def forward(self, x):
        out1 = self.pool(x)
        out2 = self.conv1(out1)
        out3 = self.conv2(out2)
        out4 = self.conv1x1(out3)
        return out4 + out1


def crop_tensor(tensor, target):
    """ Crop tensor to target size """
    _, _, tensor_depth, tensor_height, tensor_width = tensor.size()
    _, _, crop_depth, crop_height, crop_width = target.size()
    left = (tensor_width - crop_width) // 2
    top = (tensor_height - crop_height) // 2
    right = left + crop_width
    bottom = top + crop_height
    d_start = (tensor_depth - crop_depth) // 2
    d_end = d_start + crop_depth
    cropped_tensor = tensor[:, :, d_start:d_end, top: bottom, left: right]
    return cropped_tensor


class UpBlock(nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.conv1 = nn.Sequential(
            nn.ConvTranspose3d(in_channels, in_channels,
                               kernel_size=2, stride=2, padding=1),
            nn.ReLU(),
            nn.GroupNorm(32, in_channels)
        )
        self.conv2 = nn.Sequential(
            nn.Conv3d(in_channels, in_channels,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(32, in_channels)
        )
        self.conv3 = nn.Sequential(
            nn.Conv3d(in_channels, in_channels,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(32, in_channels)
        )

    def forward(self, x, down_out):
        out = self.conv1(x)
        cropped = crop_tensor(down_out, out)
        out = cropped + out # residual
        out = self.conv2(out)
        out = self.conv3(out)
        return out



class UNet3D(nn.Module):
    def __init__(self, num_classes, im_channels=3):
        super().__init__()
        self.conv_in = nn.Sequential(
            nn.Conv3d(im_channels, 64, kernel_size=3, padding=0),
            nn.ReLU(),
            nn.GroupNorm(32, 64),
            nn.Conv3d(64, 64, kernel_size=3, padding=0),
            nn.ReLU(),
            nn.GroupNorm(32, 64)
        )
        self.down1 = DownBlock(64)
        self.down2 = DownBlock(64)
        self.down3 = DownBlock(64)
        self.down4 = DownBlock(64)
        self.up1 = UpBlock(64)
        self.up2 = UpBlock(64)
        self.up3 = UpBlock(64)
        self.up4 = UpBlock(64)
        self.conv_out = nn.Sequential(
            nn.Conv3d(64, 2 * num_classes, kernel_size=1, padding=0),
            nn.ReLU()
        )

    def forward(self, x):
        out1 = self.conv_in(x)
        out2 = self.down1(out1)
        out3 = self.down2(out2)
        out4 = self.down3(out3)
        out5 = self.down4(out4)
        out = self.up1(out5, out4)
        out = self.up2(out, out3)
        out = self.up3(out, out2)
        out = self.up4(out, out1)
        out = self.conv_out(out)
        return out


class DownBlockSmall(nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.pool = nn.MaxPool3d(2)
        self.conv1 = nn.Sequential(
            nn.Conv3d(in_channels, in_channels*2,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(1, in_channels*2)
        )
        self.conv2 = nn.Sequential(
            nn.Conv3d(in_channels*2, in_channels*2,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(1, in_channels*2)
        )
        self.conv1x1 = nn.Sequential(
            # down sample channels again.
            nn.Conv3d(in_channels*2, in_channels,
                      kernel_size=1, stride=1, bias=False)
        )

    def forward(self, x):
        out1 = self.pool(x)
        out2 = self.conv1(out1)
        out3 = self.conv2(out2)
        out4 = self.conv1x1(out3)
        return out4 + out1


class UpBlockSmall(nn.Module):
    def __init__(self, in_channels):
        super().__init__()
        self.conv1 = nn.Sequential(
            nn.ConvTranspose3d(in_channels, in_channels,
                               kernel_size=2, stride=2, padding=1),
            nn.ReLU(),
            nn.GroupNorm(1, in_channels)
        )
        self.conv2 = nn.Sequential(
            nn.Conv3d(in_channels, in_channels,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(1, in_channels)
        )
        self.conv3 = nn.Sequential(
            nn.Conv3d(in_channels, in_channels,
                      kernel_size=3, padding=1),
            nn.ReLU(),
            nn.GroupNorm(1, in_channels)
        )

    def forward(self, x, down_out):
        out = self.conv1(x)
        cropped = crop_tensor(down_out, out)
        out = cropped + out # residual
        out = self.conv2(out)
        out = self.conv3(out)
        return out


class SmallUNet3D(nn.Module):
    def __init__(self, num_classes, im_channels=3):
        super().__init__()
        self.conv_in = nn.Sequential(
            nn.Conv3d(im_channels, 3, kernel_size=3, padding=0),
            nn.ReLU(),
            nn.GroupNorm(3, 3),
            nn.Conv3d(3, 3, kernel_size=3, padding=0),
            nn.ReLU(),
            nn.GroupNorm(3, 3)
        )
        print('using small unet')
        self.down1 = DownBlockSmall(3)
        self.down2 = DownBlockSmall(3)
        self.down3 = DownBlockSmall(3)
        self.down4 = DownBlockSmall(3)
        self.up1 = UpBlockSmall(3)
        self.up2 = UpBlockSmall(3)
        self.up3 = UpBlockSmall(3)
        self.up4 = UpBlockSmall(3)
        self.conv_out = nn.Sequential(
            nn.Conv3d(3, 2 * num_classes, kernel_size=1, padding=0),
            nn.ReLU()
        )

    def forward(self, x):
        out1 = self.conv_in(x)
        out2 = self.down1(out1)
        out3 = self.down2(out2)
        out4 = self.down3(out3)
        out5 = self.down4(out4)
        out = self.up1(out5, out4)
        out = self.up2(out, out3)
        out = self.up3(out, out2)
        out = self.up4(out, out1)
        out = self.conv_out(out)
        return out


"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

#pylint: disable=I1101,C0111,W0201,R0903,E0611, R0902, R0914
import os
import time
import copy
from PyQt5 import QtWidgets
from PyQt5 import QtCore
from progress_widget import BaseProgressWidget
from instructions import send_instruction
from im_utils import is_image, getDataFromDatabase
from pathlib import Path
import json


class SegmentWatchThread(QtCore.QThread):
    """
    Runs another thread.
    """
    progress_change = QtCore.pyqtSignal(int, int)
    done = QtCore.pyqtSignal()

    def __init__(self, segment_dir, total_images):
        super().__init__()
        self.segment_dir = segment_dir
        self.total_images = total_images

    def run(self):
        while True:
            # if the segment dir contains folders, then assume each folder contains the class specific predictions.

            seg_dir_contents = os.listdir(self.segment_dir)

            if len(seg_dir_contents) and os.path.isdir(os.path.join(self.segment_dir, seg_dir_contents[0])):
                done_fnames = os.listdir(os.path.join(self.segment_dir, seg_dir_contents[0]))
            else:
                done_fnames = seg_dir_contents
            done_fnames = [f for f in done_fnames if is_image(f)]
            count = len(done_fnames)
            if count >= self.total_images:
                self.done.emit()
                break
            else:
                self.progress_change.emit(count, self.total_images)
                time.sleep(0.2)


class SegmentProgressWidget(BaseProgressWidget):
    def __init__(self):
        super().__init__('Segmenting dataset')

    def run(self, segment_dir, total_images):
        self.progress_bar.setMaximum(total_images)
        self.watch_thread = SegmentWatchThread(segment_dir, total_images)
        self.watch_thread.progress_change.connect(self.onCountChanged)
        self.watch_thread.done.connect(self.done)
        self.watch_thread.start()

class SegmentFolderWidget(QtWidgets.QWidget):

    submit = QtCore.pyqtSignal()

    def __init__(self, sync_dir, instruction_dir, classes):
        super().__init__()
        # TODO # read the yaml config
        self.input_dir = None
        self.output_dir = None
        self.selected_models = []
        self.instruction_dir = instruction_dir
        self.sync_dir = sync_dir
        self.classes = classes
        self.settings = self.get_config()

        self.initUI()

    def get_config(self):
        settings_path = os.path.join(Path.home(), 'root_painter_settings.json')
        settings = json.load(open(settings_path, 'r'))
        return settings

    def segment_folder(self):
        selected_models = self.selected_models
        input_dir = self.input_dir
        output_dir = self.output_dir
      #  all_fnames = os.listdir(str(input_dir))
        all_fnames = getDataFromDatabase(self.settings)
        all_fnames = [f for f in all_fnames if is_image(f)]

        seg_classes = copy.deepcopy(self.classes)

        # need to make sure all train photos are copied now.
        content = {
            "model_paths": selected_models,
            "dataset_dir": input_dir,
            "seg_dir": output_dir,
            "file_names": all_fnames,
            "classes": seg_classes
        }
        send_instruction('segment', content, self.instruction_dir, self.sync_dir)
        self.progress_widget = SegmentProgressWidget()

        self.progress_widget.run(output_dir, len(all_fnames))
        self.progress_widget.show()
        self.close()

    def initUI(self):
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        self.setWindowTitle("Segment Folder")

        # Add specify image directory button
        in_dir_label = QtWidgets.QLabel()
        in_dir_label.setText("Input directory: Not yet specified")
        layout.addWidget(in_dir_label)
        self.in_dir_label = in_dir_label

        specify_input_dir_btn = QtWidgets.QPushButton('Specify input directory')
        specify_input_dir_btn.clicked.connect(self.select_input_dir)
        layout.addWidget(specify_input_dir_btn)

        out_dir_label = QtWidgets.QLabel()
        out_dir_label.setText("Output directory: Not yet specified")
        layout.addWidget(out_dir_label)
        self.out_dir_label = out_dir_label

        specify_output_dir_btn = QtWidgets.QPushButton('Specify output directory')
        specify_output_dir_btn.clicked.connect(self.select_output_dir)
        layout.addWidget(specify_output_dir_btn)


        model_label = QtWidgets.QLabel()
        model_label.setText("Model file: Not yet specified")
        layout.addWidget(model_label)
        self.model_label = model_label
        specify_model_btn = QtWidgets.QPushButton('Specify model file')
        specify_model_btn.clicked.connect(self.select_model)
        layout.addWidget(specify_model_btn)

        info_label = QtWidgets.QLabel()
        info_label.setText("Input directory, output directory and model"
                           " must be specified to segment folder.")
        layout.addWidget(info_label)
        self.info_label = info_label

        # Add segment button
        submit_btn = QtWidgets.QPushButton('Segment')
        submit_btn.clicked.connect(self.segment_folder)
        layout.addWidget(submit_btn)
        submit_btn.setEnabled(False)
        self.submit_btn = submit_btn


    def validate(self):
        if not self.input_dir:
            self.info_label.setText("Input directory must be specified to create project")
            self.submit_btn.setEnabled(False)
            return

        if not self.output_dir:
            self.info_label.setText("Output directory must be specified to create project")
            self.submit_btn.setEnabled(False)
            return

        if not self.selected_models:
            self.info_label.setText("Starting model must be specified to create project")
            self.submit_btn.setEnabled(False)
            return
        self.info_label.setText("")
        self.submit_btn.setEnabled(True)

    def try_submit(self):
        self.submit.emit()

    def select_input_dir(self):
        self.input_dialog = QtWidgets.QFileDialog(self)
        self.input_dialog.setFileMode(QtWidgets.QFileDialog.Directory)

        def output_selected():
            self.input_dir = self.input_dialog.selectedFiles()[0]
            self.in_dir_label.setText('Input directory: ' + self.input_dir)
            self.validate()
        self.input_dialog.fileSelected.connect(output_selected)
        self.input_dialog.open()

    def select_output_dir(self):
        self.output_dialog = QtWidgets.QFileDialog(self)
        self.output_dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        def output_selected():
            self.output_dir = self.output_dialog.selectedFiles()[0]
            self.out_dir_label.setText('Output directory: ' + self.output_dir)
            self.validate()
        self.output_dialog.fileSelected.connect(output_selected)
        self.output_dialog.open()


    def select_model(self):
        options = QtWidgets.QFileDialog.Options()
        file_paths, _ = QtWidgets.QFileDialog.getOpenFileNames(self,
                                                               "Specify model file", "",
                                                               "Pickle Files (*.pkl)",
                                                               options=options)
        if file_paths:
            file_paths = [os.path.abspath(f) for f in file_paths]
            if len(file_paths) == 1:
                self.model_label.setText('Model file: ' + file_paths[0])
            else:
                self.model_label.setText(f'{len(file_paths)} model files selected')
            self.selected_models = file_paths
            self.validate()



"""
Copyright (C) 2022 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os

def fix_config_paths(sync_dir, old_config):
    """ get paths relative to local machine """
    new_config = {}
    for k, v in old_config.items():
        if k in ['file_names',
                'file_name']:
            # names dont need a path appending
            new_config[k] = v
        elif k == 'classes':
            # classes should not be altered
            new_config[k] = v
        elif k == 'dataset_dir':
            new_config['dataset_dir'] = old_config['dataset_dir']
        elif isinstance(v, list):
            # if its a list fix each string in the list.
            new_list = []
            for e in v:
                new_val = e.replace('\\', '/')
                new_val = os.path.join(sync_dir,
                                    os.path.normpath(new_val))
                new_list.append(new_val)
            new_config[k] = new_list
        elif isinstance(v, str):
            v = v.replace('\\', '/')
            new_config[k] = os.path.join(sync_dir,
                                        os.path.normpath(v))
        else:
            new_config[k] = v
    return new_config




"""
Computing metrics on output segmentations for root images

Copyright (C) 2019, 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# pylint: disable=C0111,R0913
from datetime import datetime
from collections import namedtuple
import numpy as np

def get_metrics_str(all_metrics, to_use=None):
    out_str = ""
    for name, val in all_metrics.items():
        if to_use is None or name in to_use:
            out_str += f" {name} {val:.4g}"
    return out_str

def get_metric_csv_row(metrics):
    now_str = datetime.now().strftime('%Y-%m-%d-%H:%M:%S')
    parts = [now_str, metrics['tp'], metrics['fp'], metrics['tn'],
             metrics['fn'], round(metrics['precision'], 4),
             round(metrics['recall'], 4), round(metrics['dice'], 4)]
    return ','.join([str(p) for p in parts]) + '\n'


def get_metrics_from_arrays(y_pred, y_true):
    y_true = y_true.reshape(-1)
    y_pred = y_pred.reshape(-1)
    assert len(y_true) == len(y_pred)
    tp = np.sum(np.logical_and(y_pred == 1, y_true == 1))
    tn = np.sum(np.logical_and(y_pred == 0, y_true == 0))
    fp = np.sum(np.logical_and(y_pred == 1, y_true == 0))
    fn = np.sum(np.logical_and(y_pred == 0, y_true == 1))
    m = get_metrics(tp, fp, tn, fn)
    return m

def get_metrics(tp:int, fp:int, tn:int, fn:int) -> dict:
    # for the mtrics function in model utils
    assert not np.isnan(tp)
    assert not np.isnan(fp)
    assert not np.isnan(tn)
    assert not np.isnan(fn)
    total = (tp + tn + fp + fn)
    accuracy = (tp + tn) / total

    if tp > 0:
        precision = tp / (tp + fp)
        recall = tp / (tp + fn)
        f1 = 2 * ((precision * recall) / (precision + recall))
    else: 
        precision = recall = f1 = float('NaN')
    return {
        "accuracy": accuracy,
        "tp": tp,
        "fp": fp,
        "tn": tn,
        "fn": fn,
        "precision": precision,
        "recall": recall,
        "dice": f1,
        "true_mean": (tp + fn) / total,
        "true": (tp + fn),
        "pred": (fp + tp)
    }



def metrics_from_val_tile_refs(val_tile_refs):
    tps = 0 
    fps = 0 
    tns = 0
    fns = 0
    for ref in val_tile_refs:
        assert ref[3] is not None, ref
        (tp, fp, tn, fn) = ref[3]
        tps += tp
        fps += fp
        tns += tn
        fns += fn
    return get_metrics(tps, fps, tns, fns)



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

#pylint: disable=I1101,C0111,W0201,R0903,E0611, R0902, R0914
import os
import random
import time
import shutil
import json

from pathlib import Path, PurePosixPath
from PyQt5 import QtWidgets
from PyQt5 import QtCore

from im_utils import is_image, get_recursive_files, getDataFromDatabase
from palette import PaletteEditWidget
from name_edit_widget import NameEditWidget

class CreateProjectWidget(QtWidgets.QWidget):

    created = QtCore.pyqtSignal(Path)

    def __init__(self, sync_dir):
        super().__init__()
        self.selected_dir = None
        self.proj_name = None
        self.selected_model = None
        self.use_random_weights = True
        self.sync_dir = sync_dir
        self.settings = self.get_config()
        self.initUI()

    def get_config(self):
        settings_path = os.path.join(Path.home(), 'root_painter_settings.json')
        settings = json.load(open(settings_path, 'r'))
        return settings

    def initUI(self):
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.name_edit_widget = NameEditWidget('Project')
        self.name_edit_widget.changed.connect(self.validate)
        self.layout.addWidget(self.name_edit_widget)

        self.add_im_dir_widget()
        self.add_radio_widget()
        self.add_model_btn()
        self.add_palette_widget()
        
        self.add_info_label()
        self.add_create_btn()

    
    def add_info_label(self):
        info_label = QtWidgets.QLabel()
        info_label.setText("Name, directory and model must be specified"
                           " to create project.")
        self.layout.addWidget(info_label)
        self.info_label = info_label

    def add_im_dir_widget(self):
        # Add specify image directory button
        directory_label = QtWidgets.QLabel()
        directory_label.setText("Image directory: Not yet specified")
        self.layout.addWidget(directory_label)
        self.directory_label = directory_label

        specify_image_dir_btn = QtWidgets.QPushButton('Specify image directory')
        specify_image_dir_btn.clicked.connect(self.select_photo_dir)
        self.layout.addWidget(specify_image_dir_btn)

    def add_radio_widget(self):
        radio_widget = QtWidgets.QWidget()
        radio_layout = QtWidgets.QHBoxLayout()
        radio_widget.setLayout(radio_layout)
        self.layout.addWidget(radio_widget)

        # Add radio, use random weight or specify model file.
        radio = QtWidgets.QRadioButton("Random Weights")
        radio.setChecked(True)
        radio.name = "random"
        radio.toggled.connect(self.on_radio_clicked)
        radio_layout.addWidget(radio)

        radio = QtWidgets.QRadioButton("Specify Model")
        radio.name = "specify"
        radio.toggled.connect(self.on_radio_clicked)
        radio_layout.addWidget(radio)

    def add_model_btn(self):
        model_label = QtWidgets.QLabel()
        model_label.setText("Model: Please specify model file")
        self.layout.addWidget(model_label)
        self.model_label = model_label
        specify_model_btn = QtWidgets.QPushButton('Specify model file')
        specify_model_btn.clicked.connect(self.select_model)
        self.specify_model_btn = specify_model_btn
        self.layout.addWidget(specify_model_btn)

        self.model_label.setVisible(False)
        self.specify_model_btn.setVisible(False)

    def add_palette_widget(self):
        self.palette_edit_widget = PaletteEditWidget()
        self.palette_edit_widget.changed.connect(self.validate)
        self.layout.addWidget(self.palette_edit_widget)

    def add_create_btn(self):
        # Add create button
        create_project_btn = QtWidgets.QPushButton('Create project')
        create_project_btn.clicked.connect(self.create_project)
        self.layout.addWidget(create_project_btn)
        create_project_btn.setEnabled(False)
        self.create_project_btn = create_project_btn

    def on_radio_clicked(self):
        radio = self.sender()
        if radio.isChecked():
            print("Radio is %s" % (radio.name))
            specify = (radio.name == 'specify')
            self.model_label.setVisible(specify)
            self.specify_model_btn.setVisible(specify)
            self.use_random_weights = not specify
            self.validate()

    def validate(self):
        self.proj_name = self.name_edit_widget.name
        if not self.proj_name:
            self.info_label.setText("Name must be specified to create project")
            self.create_project_btn.setEnabled(False)
            return

        if not self.selected_dir:
            self.info_label.setText("Directory must be specified to create project")
            self.create_project_btn.setEnabled(False)
            return

        if not self.use_random_weights and not self.selected_model:
            self.info_label.setText("Starting model must be specified to create project")
            self.create_project_btn.setEnabled(False)
            return

        #cur_files = os.listdir(self.selected_dir)
        # TODO get files using db
        #cur_files = get_recursive_files(self.selected_dir)
        cur_files = getDataFromDatabase(self.settings)
        cur_files = [f for f in cur_files if is_image(f)]

        if not cur_files:
            message = "Folder contains no compatible images. Valid formats include .nifti, .nrrd and dicom"
            self.info_label.setText(message)
            self.create_project_btn.setEnabled(False)
            return

        if len(self.palette_edit_widget.get_brush_data()) < 1:
            self.info_label.setText(f"At least one foreground brush must be specified")
            self.create_project_btn.setEnabled(False)
            return

        self.project_location = os.path.join('projects', self.proj_name)
        if os.path.exists(os.path.join(self.sync_dir, self.project_location)):
            self.info_label.setText(f"Project with name {self.proj_name} already exists")
            self.create_project_btn.setEnabled(False)
        else:
            self.info_label.setText(f"Project location: {self.project_location}")
            self.create_project_btn.setEnabled(True)

    def select_photo_dir(self):
        self.photo_dialog = QtWidgets.QFileDialog(self)
        self.photo_dialog.setFileMode(QtWidgets.QFileDialog.Directory)
        def output_selected():
            self.selected_dir = self.photo_dialog.selectedFiles()[0]
            self.directory_label.setText('Image directory: ' + self.selected_dir)
            self.validate()

        self.photo_dialog.fileSelected.connect(output_selected)
        self.photo_dialog.open()

    def select_model(self):
        options = QtWidgets.QFileDialog.Options()
        file_path, _ = QtWidgets.QFileDialog.getOpenFileName(self,
                                                             "Specify model file", "",
                                                             "Pickle Files (*.pkl)",
                                                             options=options)
        if file_path:
            file_path = os.path.abspath(file_path)
            self.selected_model = file_path
            self.model_label.setText('Model file: ' + self.selected_model)
            self.validate()

    def create_project(self):
        project_name = self.proj_name
        project_location = Path(self.project_location)

        classes = self.palette_edit_widget.get_brush_data()

        dataset_path = os.path.abspath(self.selected_dir)
        datasets_dir = str(self.sync_dir / 'datasets')
    
        # if not dataset_path.startswith(datasets_dir):
        #     message = ("When creating a project the selected dataset must be in "
        #                "the datasets folder. The selected dataset is "
        #                f"{dataset_path} and the datasets folder is "
        #                f"{datasets_dir}.")
        #     QtWidgets.QMessageBox.about(self, 'Project Creation Error', message)
        #     return

        os.makedirs(self.sync_dir / project_location)
        proj_file_path = (self.sync_dir / project_location /
                          (project_name + '.seg_proj'))

        if len(classes) > 1:
            # Create annotation folder for each class
            for class_name in classes:
                os.makedirs(self.sync_dir / project_location /
                            'annotations' / class_name / 'train')
                os.makedirs(self.sync_dir / project_location /
                            'annotations' / class_name / 'val')
                os.makedirs(self.sync_dir / project_location /
                            'segmentations' / class_name)
                os.makedirs(self.sync_dir / project_location /
                            'train_segmentations' / class_name)
 
        else:
            # Create only the 'annotation' and 'segmentation' folders - classic structure
            os.makedirs(self.sync_dir / project_location / 'annotations' / 'train')
            os.makedirs(self.sync_dir / project_location / 'annotations' / 'val')
            os.makedirs(self.sync_dir / project_location / 'segmentations')
            os.makedirs(self.sync_dir / project_location / 'train_segmentations')

        os.makedirs(self.sync_dir / project_location / 'models')
        os.makedirs(self.sync_dir / project_location / 'messages')
        os.makedirs(self.sync_dir / project_location / 'logs')

        if self.use_random_weights:
            original_model_file = 'random weights'
        else:
            model_num = 1
            model_name = str(model_num).zfill(6)
            model_name += '_' + str(int(round(time.time()))) + '.pkl'
            shutil.copyfile(self.selected_model,
                            self.sync_dir / project_location /
                            'models' / model_name)
            original_model_file = self.selected_model


        dataset = dataset_path
        # get files in random order for training.
        #all_fnames = os.listdir(dataset_path)
       # all_fnames = get_recursive_files(dataset_path)
        all_fnames = getDataFromDatabase(self.settings)
        ## TODO #
        # images only
        all_fnames = [a for a in all_fnames if is_image(a)]

        all_fnames = sorted(all_fnames)
        random.shuffle(all_fnames)

        # create project file.
        project_info = {
            'name': project_name,
            'dataset': dataset,
            'original_model_file': original_model_file,
            'location': str(PurePosixPath(project_location))
        }

        # only add classes info if the palette is defined.
        # otherwise the server will default to single class (fg/bg)
        if hasattr(self, 'palette_edit_widget'):
            project_info['classes'] = classes

        # add these at the end because it makes the json more readable
        # to have the short entries at the top.
        # REDUCE to basename and one level parent path
        # all_fnames_reduced = []
        # for fname_i in all_fnames:
        #     red = os.path.join(os.path.basename(os.path.dirname(fname_i)),
        #                  os.path.basename(fname_i))
        #     all_fnames_reduced.append(red)
            
      #  project_info['file_names'] = all_fnames_reduced
        project_info['file_names'] = all_fnames
        with open(proj_file_path, 'w') as json_file:
            json.dump(project_info, json_file, indent=4)
        #self.created.emit(str(proj_file_path))
        self.created.emit(proj_file_path)
        self.close()



"""
Copyright (C) 2020 Abraham George Smith

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

#pylint: disable=C0111
import os
import json
from pathlib import Path

def fix_path(path, sync_dir):
    """ fix path by removing everything that might
        be different on another machine """
    path_str = str(path)
    path_str = path_str.replace(str(sync_dir), '')
    path_str = path_str.replace('\\', '/') # server is unix.
    
    # sometimes the path has forward slash so the sync dir doesn't 
    # get removed. In this case get the sync dir with forward
    # slash and then try to remove that.
    sync_dir_forward = str(sync_dir).replace('\\', '/')
    path_str = path_str.replace(sync_dir_forward, '')
    # only remove first character if it is a forward slash
    if path_str[0] == '/':
        return path_str[1:] 
    return path_str


def fix_instruction_paths(old_config, sync_dir):
    # remove part of path that might be different on server.
    new_config = {}
    for k, v in old_config.items():
        # names dont need anything removing
        # classes dont need anything changing
        if k in ['file_names',
                'classes',
                'cient_ip',
                'client_username']:
            new_config[k] = v
        elif isinstance(v, list):
            # if its a list fix each string in the list.
            new_list = []
            for e in v:
                new_val = fix_path(e, sync_dir)
                new_list.append(new_val)
            new_config[k] = new_list
        elif isinstance(v, str):
            new_config[k] = fix_path(v, sync_dir)
            if k == 'dataset_dir':
                new_config['dataset_dir'] = old_config['dataset_dir']
        elif isinstance(v, Path):
            new_config[k] = fix_path(v, sync_dir)
            if k == 'dataset_dir':
                new_config['dataset_dir'] = old_config['dataset_dir']
        else:
            new_config[k] = v
    return new_config


def send_instruction(name, content, instruction_dir, sync_dir):
    content = fix_instruction_paths(content, sync_dir)
    # append a hash to avoid over writing older instructions that
    # have not yet finished.
    if name == 'start_training':
        content['dataset_dir'] = str(content['dataset_dir'])
        
    hash_str = '_' + str(hash(json.dumps(content)))
    fpath = os.path.join(instruction_dir, name + hash_str)
    # if this instruction already exists then don't send again
    if not os.path.isfile(fpath):
        with open(fpath, 'w') as json_file:
            json.dump(content, json_file, indent=4)



"""
Human in the loop deep learning segmentation for biological images

Copyright (C) 2020 Abraham George Smith
Copyright (C) 2022 Abraham George Smith


This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from pathlib import Path
import os
import json
import argparse
from cagpaint.trainer import Trainer
from startup import startup_setup
import yaml
# Import all the necessary submodules

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--syncdir',
                        help=('location of directory where data is'
                              ' synced between the client and server'),
                        required=True)
    parser.add_argument('--config_file',
                        help="""
                'please specify path to yaml config file""", type=str,
                required=True,
                default=False)

    settings_path = os.path.join(Path.home(), 'root_painter_settings.json')
   
    settings = None
    
    args = parser.parse_args()
    arguments = vars(args)
    with open(args.config_file, 'r') as f:
        config = yaml.safe_load(f)
    arguments['sync_dir'] = settings_path
    arguments = {**config, **arguments}
    sync_dir = args.syncdir
    startup_setup(arguments, sync_dir=sync_dir)
    import torch
    print('cuda=?' ,torch.cuda.is_available())

    if settings and 'auto_complete' in settings and settings['auto_complete']:
        ip = settings['server_ip']
        port = settings['server_port']
        trainer = Trainer(sync_dir, config, ip, port)
    else:
        trainer = Trainer(sync_dir, config)

    trainer.main_loop()
